# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

from .. core._core_utils import add_method
from .. import _otio


@add_method(_otio.Transition)
def __str__(self):
    return 'Transition("{}", "{}", {}, {}, {})'.format(
        self.name,
        self.transition_type,
        self.in_offset,
        self.out_offset,
        self.metadata
    )


@add_method(_otio.Transition)
def __repr__(self):
    return (
        'otio.schema.Transition('
        'name={}, '
        'transition_type={}, '
        'in_offset={}, '
        'out_offset={}, '
        'metadata={}'
        ')'.format(
            repr(self.name),
            repr(self.transition_type),
            repr(self.in_offset),
            repr(self.out_offset),
            repr(self.metadata),
        )
    )



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

from .. core._core_utils import add_method
from .. import _otio


@add_method(_otio.SerializableCollection)
def __str__(self):
    return "SerializableCollection({}, {}, {})".format(
        str(self.name),
        str(list(self)),
        str(self.metadata)
    )


@add_method(_otio.SerializableCollection)
def __repr__(self):
    return (
        "otio.{}("
        "name={}, "
        "children={}, "
        "metadata={}"
        ")".format(
            "schema.SerializableCollection",
            repr(self.name),
            repr(list(self)),
            repr(self.metadata)
        )
    )



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

from .. core._core_utils import add_method
from .. import _otio


@add_method(_otio.GeneratorReference)
def __str__(self):
    return 'GeneratorReference("{}", "{}", {}, {}, {})'.format(
        self.name,
        self.generator_kind,
        self.parameters,
        self.available_image_bounds,
        self.metadata
    )


@add_method(_otio.GeneratorReference)
def __repr__(self):
    return (
        'otio.schema.GeneratorReference('
        'name={}, '
        'generator_kind={}, '
        'parameters={}, '
        'available_image_bounds={}, '
        'metadata={}'
        ')'.format(
            repr(self.name),
            repr(self.generator_kind),
            repr(self.parameters),
            repr(self.available_image_bounds),
            repr(self.metadata),
        )
    )



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

from . _core_utils import add_method
from .. import _otio


@add_method(_otio.Composable)
def __repr__(self):
    return (
        "otio.{}("
        "name={}, "
        "metadata={}"
        ")".format(
            "core.Composable",
            repr(str(self.name)),
            repr(self.metadata)
        )
    )


@add_method(_otio.Composable)
def __str__(self):
    return "{}({}, {})".format(
        "Composable",
        str(self.name),
        str(self.metadata)
    )



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

from . _opentime import ( # noqa
    RationalTime,
    TimeRange,
    TimeTransform,
)

__all__ = [
    'RationalTime',
    'TimeRange',
    'TimeTransform',
    'from_frames',
    'from_timecode',
    'from_time_string',
    'from_seconds',
    'to_timecode',
    'to_nearest_timecode',
    'to_frames',
    'to_seconds',
    'to_time_string',
    'range_from_start_end_time',
    'range_from_start_end_time_inclusive',
    'duration_from_start_end_time',
    'duration_from_start_end_time_inclusive',
]

from_frames = RationalTime.from_frames
from_timecode = RationalTime.from_timecode
from_time_string = RationalTime.from_time_string
from_seconds = RationalTime.from_seconds

range_from_start_end_time = TimeRange.range_from_start_end_time
range_from_start_end_time_inclusive = TimeRange.range_from_start_end_time_inclusive
duration_from_start_end_time = RationalTime.duration_from_start_end_time
duration_from_start_end_time_inclusive = (
    RationalTime.duration_from_start_end_time_inclusive
)


def to_timecode(rt, rate=None, drop_frame=None):
    """Convert a :class:`~RationalTime` into a timecode string."""
    return (
        rt.to_timecode()
        if rate is None and drop_frame is None
        else rt.to_timecode(rate, drop_frame)
    )


def to_nearest_timecode(rt, rate=None, drop_frame=None):
    """Convert a :class:`~RationalTime` into a timecode string."""
    return (
        rt.to_nearest_timecode()
        if rate is None and drop_frame is None
        else rt.to_nearest_timecode(rate, drop_frame)
    )


def to_frames(rt, rate=None):
    """Turn a :class:`~RationalTime` into a frame number."""
    return rt.to_frames() if rate is None else rt.to_frames(rate)


def to_seconds(rt):
    """Convert a :class:`~RationalTime` into float seconds"""
    return rt.to_seconds()


def to_time_string(rt):
    """
    Convert this timecode to time as used by ffmpeg, formatted as
    ``hh:mm:ss`` where ss is an integer or decimal number.
    """
    return rt.to_time_string()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Utilities for conversion between urls and file paths"""

import os
import urllib
from urllib import request
import pathlib


def url_from_filepath(fpath):
    """Convert a filesystem path to an url in a portable way.

    ensures that `fpath` conforms to the following pattern:
        * if it is an absolute path, "file:///path/to/thing"
        * if it is a relative path, "path/to/thing"

    In other words, if you pass in:
        * "/var/tmp/thing.otio" -> "file:///var/tmp/thing.otio"
        * "subdir/thing.otio" -> "tmp/thing.otio"
    """

    try:
        # appears to handle absolute windows paths better, which are absolute
        # and start with a drive letter.
        return urllib.parse.unquote(pathlib.PurePath(fpath).as_uri())
    except ValueError:
        # scheme is "file" for absolute paths, else ""
        scheme = "file" if os.path.isabs(fpath) else ""

        # handles relative paths
        return urllib.parse.urlunparse(
            urllib.parse.ParseResult(
                scheme=scheme,
                path=fpath,
                netloc="",
                params="",
                query="",
                fragment=""
            )
        )


def filepath_from_url(urlstr):
    """
    Take an url and return a filepath.

    URLs can either be encoded according to the `RFC 3986`_ standard or not.
    Additionally, Windows mapped drive letter and UNC paths need to be
    accounted for when processing URL(s); however, there are `ongoing
    discussions`_ about how to best handle this within Python developer
    community. This function is meant to cover these scenarios in the interim.

    .. _RFC 3986: https://tools.ietf.org/html/rfc3986#section-2.1
    .. _ongoing discussions: https://discuss.python.org/t/file-uris-in-python/15600
    """

    # Parse provided URL
    parsed_result = urllib.parse.urlparse(urlstr)

    # De-encode the parsed path
    decoded_parsed_path = urllib.parse.unquote(parsed_result.path)

    # Convert the parsed URL to a path
    filepath = pathlib.PurePath(
        request.url2pathname(decoded_parsed_path)
    )

    # If the network location is a window drive, reassemble the path
    if pathlib.PureWindowsPath(parsed_result.netloc).drive:
        filepath = pathlib.PurePath(parsed_result.netloc + decoded_parsed_path)

    # If the specified index is a windows drive, then append it to the other
    # parts
    elif pathlib.PureWindowsPath(filepath.parts[0]).drive:
        filepath = pathlib.PurePosixPath(filepath.drive, *filepath.parts[1:])

    # If the specified index is a windows drive, then offset the path
    elif (
        # relative paths may not have a parts[1]
        len(filepath.parts) > 1
        and pathlib.PureWindowsPath(filepath.parts[1]).drive
    ):
        # Remove leading "/" if/when `request.url2pathname` yields
        # "/S:/path/file.ext"
        filepath = pathlib.PurePosixPath(*filepath.parts[1:])

    # Should catch UNC paths,
    # as parsing "file:///some/path/to/file.ext" doesn't provide a netloc
    elif parsed_result.netloc and parsed_result.netloc != 'localhost':
        # Paths of type: "file://host/share/path/to/file.ext" provide "host" as
        # netloc
        filepath = pathlib.PurePath(
            '//',
            parsed_result.netloc + decoded_parsed_path
        )

        # Executing `as_posix` on Windows seems to generate a path with only 1
        # leading `/`, so we insert another `/` at the front of the string path
        # to match Linux and Windows UNC conventions and return it.
        conformed_filepath = filepath.as_posix()
        if not conformed_filepath.startswith('//'):
            conformed_filepath = '/' + conformed_filepath
        return conformed_filepath

    # Convert "\" to "/" if needed
    return filepath.as_posix()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Adapter for reading and writing native .otio json files."""

from .. import (
    core,
    versioning,
    exceptions
)

import os

_DEFAULT_VERSION_ENVVAR = "OTIO_DEFAULT_TARGET_VERSION_FAMILY_LABEL"


def read_from_file(filepath):
    """
    De-serializes an OpenTimelineIO object from a file

    Args:
        filepath (str): The path to an otio file to read from

    Returns:
        OpenTimeline: An OpenTimeline object
    """
    return core.deserialize_json_from_file(filepath)


def read_from_string(input_str):
    """
    De-serializes an OpenTimelineIO object from a json string

    Args:
        input_str (str): A string containing json serialized otio contents

    Returns:
        OpenTimeline: An OpenTimeline object
    """
    return core.deserialize_json_from_string(input_str)


def _fetch_downgrade_map_from_env():
    version_envvar = os.environ[_DEFAULT_VERSION_ENVVAR]

    try:
        family, label = version_envvar.split(":")
    except ValueError:
        raise exceptions.InvalidEnvironmentVariableError(
            "Environment variable '{}' is incorrectly formatted with '{}'."
            "Variable must be formatted as 'FAMILY:LABEL'".format(
                _DEFAULT_VERSION_ENVVAR,
                version_envvar,
            )
        )

    try:
        # technically fetch_map returns an AnyDictionary, but the pybind11
        # code wrapping the call to the serializer expects a python
        # dictionary.  This turns it back into a normal dictionary.
        return dict(versioning.fetch_map(family, label))
    except KeyError:
        raise exceptions.InvalidEnvironmentVariableError(
            "Environment variable '{}' is requesting family '{}' and label"
            " '{}', however this combination does not exist in the "
            "currently loaded manifests.  Full version map: {}".format(
                _DEFAULT_VERSION_ENVVAR,
                family,
                label,
                versioning.full_map()
            )
        )


def write_to_string(input_otio, target_schema_versions=None, indent=4):
    """
    Serializes an OpenTimelineIO object into a string

    Args:
        input_otio (OpenTimeline): An OpenTimeline object
        indent (int): number of spaces for each json indentation level. Use\
            -1 for no indentation or newlines.

    If target_schema_versions is None and the environment variable
    "OTIO_DEFAULT_TARGET_VERSION_FAMILY_LABEL" is set, will read a map out of
    that for downgrade target.  The variable should be of the form
    FAMILY:LABEL, for example "MYSTUDIO:JUNE2022".

    Returns:
        str: A json serialized string representation

    Raises:
        otio.exceptions.InvalidEnvironmentVariableError: if there is a problem
        with the default environment variable
        "OTIO_DEFAULT_TARGET_VERSION_FAMILY_LABEL".
    """

    if (
            target_schema_versions is None
            and _DEFAULT_VERSION_ENVVAR in os.environ
    ):
        target_schema_versions = _fetch_downgrade_map_from_env()

    return core.serialize_json_to_string(
        input_otio,
        target_schema_versions,
        indent
    )


def write_to_file(
        input_otio,
        filepath,
        target_schema_versions=None,
        indent=4
):
    """
    Serializes an OpenTimelineIO object into a file

    Args:

        input_otio (OpenTimeline): An OpenTimeline object
        filepath (str): The name of an otio file to write to
        indent (int): number of spaces for each json indentation level.\
            Use -1 for no indentation or newlines.

    If target_schema_versions is None and the environment variable
    "OTIO_DEFAULT_TARGET_VERSION_FAMILY_LABEL" is set, will read a map out of
    that for downgrade target.  The variable should be of the form
    FAMILY:LABEL, for example "MYSTUDIO:JUNE2022".

    Returns:
        bool: Write success

    Raises:
        ValueError: on write error
        otio.exceptions.InvalidEnvironmentVariableError: if there is a problem
        with the default environment variable
        "OTIO_DEFAULT_TARGET_VERSION_FAMILY_LABEL".
    """

    if (
        target_schema_versions is None
        and _DEFAULT_VERSION_ENVVAR in os.environ
    ):
        target_schema_versions = _fetch_downgrade_map_from_env()

    return core.serialize_json_to_file(
        input_otio,
        filepath,
        target_schema_versions,
        indent
    )



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Implementation of the OTIO internal `Adapter` system.

For information on writing adapters, please consult:
https://opentimelineio.readthedocs.io/en/latest/tutorials/write-an-adapter.html # noqa
"""

import inspect
import collections
import copy
from typing import List

from .. import (
    core,
    plugins,
    media_linker,
    hooks,
)


@core.register_type
class Adapter(plugins.PythonPlugin):
    """Adapters convert between OTIO and other formats.

    Note that this class is not subclassed by adapters. Rather, an adapter is
    a python module that implements at least one of the following functions:

    .. code-block:: python

        write_to_string(input_otio)
        write_to_file(input_otio, filepath) (optionally inferred)
        read_from_string(input_str)
        read_from_file(filepath) (optionally inferred)

    ...as well as a small json file that advertises the features of the adapter
    to OTIO.  This class serves as the wrapper around these modules internal
    to OTIO.  You should not need to extend this class to create new adapters
    for OTIO.

    For more information: https://opentimelineio.readthedocs.io/en/latest/tutorials/write-an-adapter.html. # noqa
    """
    _serializable_label = "Adapter.1"

    def __init__(
        self,
        name=None,
        filepath=None,
        suffixes=None
    ):
        plugins.PythonPlugin.__init__(
            self,
            name,
            filepath
        )

        self.suffixes = suffixes or []

    suffixes = core.serializable_field(
        "suffixes",
        type([]),
        doc="File suffixes associated with this adapter."
    )

    def has_feature(self, feature_string):
        """
        return true if adapter supports feature_string, which must be a key
        of the _FEATURE_MAP dictionary.

        Will trigger a call to :meth:`.PythonPlugin.module`, which imports the plugin.
        """

        if feature_string.lower() not in _FEATURE_MAP:
            return False

        search_strs = _FEATURE_MAP[feature_string]

        try:
            return any(hasattr(self.module(), s) for s in search_strs)
        except ImportError:
            # @TODO: should issue a warning that the plugin was not importable?
            return False

    def read_from_file(
        self,
        filepath,
        media_linker_name=media_linker.MediaLinkingPolicy.ForceDefaultLinker,
        media_linker_argument_map=None,
        hook_function_argument_map=None,
        **adapter_argument_map
    ):
        """Execute the read_from_file function on this adapter.

        If read_from_string exists, but not read_from_file, execute that with
        a trivial file object wrapper.
        """

        media_linker_argument_map = copy.deepcopy(
            media_linker_argument_map or {}
        )

        hook_function_argument_map = copy.deepcopy(
            hook_function_argument_map or {}
        )
        hook_function_argument_map['adapter_arguments'] = copy.deepcopy(
            adapter_argument_map
        )
        hook_function_argument_map['media_linker_argument_map'] = (
            media_linker_argument_map
        )

        if self.has_feature("hooks"):
            adapter_argument_map[
                "hook_function_argument_map"
            ] = hook_function_argument_map

        result = None

        if (
            not self.has_feature("read_from_file") and
            self.has_feature("read_from_string")
        ):
            with open(filepath, encoding="utf-8") as fo:
                contents = fo.read()
            result = self._execute_function(
                "read_from_string",
                input_str=contents,
                **adapter_argument_map
            )
        else:
            result = self._execute_function(
                "read_from_file",
                filepath=filepath,
                **adapter_argument_map
            )

        result = hooks.run(
            "post_adapter_read",
            result,
            extra_args=hook_function_argument_map
        )

        if media_linker_name and (
            media_linker_name != media_linker.MediaLinkingPolicy.DoNotLinkMedia
        ):
            _with_linked_media_references(
                result,
                media_linker_name,
                media_linker_argument_map
            )

        result = hooks.run(
            "post_media_linker",
            result,
            extra_args=media_linker_argument_map
        )

        return result

    def write_to_file(
        self,
        input_otio,
        filepath,
        hook_function_argument_map=None,
        **adapter_argument_map
    ):
        """Execute the write_to_file function on this adapter.

        If write_to_string exists, but not write_to_file, execute that with
        a trivial file object wrapper.
        """

        hook_function_argument_map = copy.deepcopy(
            hook_function_argument_map or {}
        )
        hook_function_argument_map['adapter_arguments'] = copy.deepcopy(
            adapter_argument_map
        )

        # Store file path for use in hooks
        hook_function_argument_map['_filepath'] = filepath

        if self.has_feature("hooks"):
            adapter_argument_map[
                "hook_function_argument_map"
            ] = hook_function_argument_map

        input_otio = hooks.run("pre_adapter_write", input_otio,
                               extra_args=hook_function_argument_map)
        if (
            not self.has_feature("write_to_file") and
            self.has_feature("write_to_string")
        ):
            result = self.write_to_string(input_otio, **adapter_argument_map)
            with open(filepath, 'w', encoding="utf-8") as fo:
                fo.write(result)
            result = filepath

        else:
            result = self._execute_function(
                "write_to_file",
                input_otio=input_otio,
                filepath=filepath,
                **adapter_argument_map
            )

        hooks.run(
            "post_adapter_write",
            input_otio,
            extra_args=hook_function_argument_map
        )

        return result

    def read_from_string(
        self,
        input_str,
        media_linker_name=media_linker.MediaLinkingPolicy.ForceDefaultLinker,
        media_linker_argument_map=None,
        hook_function_argument_map=None,
        **adapter_argument_map
    ):
        """Call the read_from_string function on this adapter."""
        hook_function_argument_map = copy.deepcopy(
            hook_function_argument_map or {}
        )
        hook_function_argument_map['adapter_arguments'] = copy.deepcopy(
            adapter_argument_map
        )
        hook_function_argument_map['media_linker_argument_map'] = copy.deepcopy(
            media_linker_argument_map
        )

        if self.has_feature("hooks"):
            adapter_argument_map[
                "hook_function_argument_map"
            ] = hook_function_argument_map

        result = self._execute_function(
            "read_from_string",
            input_str=input_str,
            **adapter_argument_map
        )

        result = hooks.run(
            "post_adapter_read",
            result,
            extra_args=hook_function_argument_map
        )

        if media_linker_name and (
            media_linker_name != media_linker.MediaLinkingPolicy.DoNotLinkMedia
        ):
            _with_linked_media_references(
                result,
                media_linker_name,
                media_linker_argument_map
            )

        # @TODO: Should this run *ONLY* if the media linker ran?
        result = hooks.run(
            "post_media_linker",
            result,
            extra_args=hook_function_argument_map
        )

        return result

    def write_to_string(
        self,
        input_otio,
        hook_function_argument_map=None,
        **adapter_argument_map
    ):
        """Call the write_to_string function on this adapter."""

        hook_function_argument_map = copy.deepcopy(
            hook_function_argument_map or {}
        )
        hook_function_argument_map['adapter_arguments'] = copy.deepcopy(
            adapter_argument_map
        )
        input_otio = hooks.run(
            "pre_adapter_write",
            input_otio,
            extra_args=hook_function_argument_map
        )

        return self._execute_function(
            "write_to_string",
            input_otio=input_otio,
            **adapter_argument_map
        )

    def adapter_hook_names(self) -> List[str]:
        """Returns a list of hooks claimed by the adapter.

        In addition to the hook being declared in the manifest, it should also be
        returned here, so it can be attributed to the adapter.
        """
        if not self.has_feature("hooks"):
            return []
        return self._execute_function("adapter_hook_names")

    def __str__(self):
        return (
            "Adapter("
            "{}, "
            "{}, "
            "{}"
            ")".format(
                repr(self.name),
                repr(self.filepath),
                repr(self.suffixes),
            )
        )

    def __repr__(self):
        return (
            "otio.adapter.Adapter("
            "name={}, "
            "filepath={}, "
            "suffixes={}"
            ")".format(
                repr(self.name),
                repr(self.filepath),
                repr(self.suffixes),
            )
        )

    def plugin_info_map(self):
        """Adds extra adapter-specific information to call to the parent fn."""

        result = super().plugin_info_map()

        features = collections.OrderedDict()
        result["supported features"] = features

        for feature in sorted(_FEATURE_MAP.keys()):
            if feature in ["read", "write", "hooks"]:
                continue

            if self.has_feature(feature):
                features[feature] = collections.OrderedDict()

                # find the function
                args = []
                for fn_name in _FEATURE_MAP[feature]:
                    if hasattr(self.module(), fn_name):
                        fn = getattr(self.module(), fn_name)
                        args = inspect.getfullargspec(fn)
                        docs = inspect.getdoc(fn)
                        break

                if args:
                    features[feature]["args"] = args.args
                    features[feature]["doc"] = docs

        # check if there are any adapter specific-hooks and get their names
        if self.has_feature("hooks"):
            adapter_hooks_names_fn = getattr(
                self.module(), _FEATURE_MAP["hooks"][0], None
            )
            if adapter_hooks_names_fn:
                features["hooks"] = adapter_hooks_names_fn()

        return result


def _with_linked_media_references(
    read_otio,
    media_linker_name,
    media_linker_argument_map
):
    """Link media references in the read_otio if possible.

    Makes changes in place and returns the read_otio structure back.
    """

    if not read_otio or not media_linker.from_name(media_linker_name):
        return read_otio

    # not every object the adapter reads has an "find_clips" method, so this
    # skips objects without one.
    clpfn = getattr(read_otio, "find_clips", None)
    if clpfn is None:
        return read_otio

    for cl in read_otio.find_clips():
        new_mr = media_linker.linked_media_reference(
            cl,
            media_linker_name,
            # @TODO: should any context get wired in at this point?
            media_linker_argument_map
        )
        if new_mr is not None:
            cl.media_reference = new_mr

    return read_otio


# map of attr to look for vs feature name in the adapter plugin
_FEATURE_MAP = {
    'read_from_file': ['read_from_file'],
    'read_from_string': ['read_from_string'],
    'read': ['read_from_file', 'read_from_string'],
    'write_to_file': ['write_to_file'],
    'write_to_string': ['write_to_string'],
    'write': ['write_to_file', 'write_to_string'],
    'hooks': ['adapter_hook_names']
}



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

import ast

from .. import (
    media_linker,
)

"""Utilities for OpenTimelineIO commandline modules."""


def arg_list_to_map(arg_list, label):
    """
    Convert an argument of the form -A foo=bar from the parsed result to a map.
    """

    argument_map = {}
    for pair in arg_list:
        if '=' not in pair:
            raise ValueError(
                "error: {} arguments must be in the form key=value"
                " got: {}".format(label, pair)
            )

        key, val = pair.split('=', 1)  # only split on the 1st '='
        try:
            # Sometimes we need to pass a bool, int, list, etc.
            parsed_value = ast.literal_eval(val)
        except (ValueError, SyntaxError):
            # Fall back to a simple string
            parsed_value = val
        argument_map[key] = parsed_value

    return argument_map


def media_linker_name(ml_name_arg):
    """
    Parse commandline arguments for the media linker, which can be not set
    (fall back to default), "" or "none" (don't link media) or the name of a
    media linker to use.
    """
    if ml_name_arg.lower() == 'default':
        media_linker_name = media_linker.MediaLinkingPolicy.ForceDefaultLinker
    elif ml_name_arg.lower() in ['none', '']:
        media_linker_name = media_linker.MediaLinkingPolicy.DoNotLinkMedia
    else:
        media_linker_name = ml_name_arg

    return media_linker_name



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

# flake8: noqa

"""User facing classes."""

from .. _otio import (
    Box2d,
    Clip,
    Effect,
    TimeEffect,
    LinearTimeWarp,
    ExternalReference,
    FreezeFrame,
    Gap,
    GeneratorReference,
    ImageSequenceReference,
    Marker,
    MissingReference,
    SerializableCollection,
    Stack,
    Timeline,
    Track,
    Transition,
    V2d,
)

MarkerColor = Marker.Color
TrackKind = Track.Kind
TransitionTypes = Transition.Type
NeighborGapPolicy = Track.NeighborGapPolicy

from . schemadef import (
    SchemaDef
)

from . import (
    box2d,
    clip,
    effect,
    external_reference,
    generator_reference,
    image_sequence_reference,
    marker,
    serializable_collection,
    timeline,
    transition,
    v2d,
)

def timeline_from_clips(clips):
    """Convenience for making a single track timeline from a list of clips."""

    trck = Track(children=clips)
    return Timeline(tracks=[trck])

__all__ = [
    'Box2d',
    'Clip',
    'Effect',
    'TimeEffect',
    'LinearTimeWarp',
    'ExternalReference',
    'FreezeFrame',
    'Gap',
    'GeneratorReference',
    'ImageSequenceReference',
    'Marker',
    'MissingReference',
    'SerializableCollection',
    'Stack',
    'Timeline',
    'Transition',
    'SchemaDef',
    'timeline_from_clips',
    'V2d'
]



#!/usr/bin/env python
#
# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Print information about the OTIO plugin ecosystem."""

import argparse
import fnmatch
import textwrap

import opentimelineio as otio

OTIO_PLUGIN_TYPES = ['all'] + otio.plugins.manifest.OTIO_PLUGIN_TYPES


def _parsed_args():
    """ parse commandline arguments with argparse """

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        '-p',
        '--plugin-types',
        type=str,
        default='all',
        nargs='+',
        choices=OTIO_PLUGIN_TYPES,
        help=(
            'Comma separated list of which kinds of plugins to print info on.'
        )
    )
    parser.add_argument(
        '-a',
        '--attribs',
        type=str,
        default=['*'],
        nargs='+',
        help=(
            'Comma separated list of globs of which attributes to print info'
            ' on.'
        )

    )
    parser.add_argument(
        '-l',
        '--long-docs',
        default=False,
        action="store_true",
        help="Print full docstring instead of just the summary line.",
    )
    parser.add_argument(
        'plugpattern',
        type=str,
        default='*',
        nargs='?',
        help='Only print information about plugins that match this glob.'
    )
    parser.add_argument(
        '--version',
        default=False,
        action="store_true",
        help=(
            "Print the otio and installed plugin package version "
            "information to the commandline."
        ),
    )

    return parser.parse_args()


def _supported_features_formatted(feature_map, _):
    if feature_map:
        print("    explicit supported features:")
    for thing, args in feature_map.items():
        # skip hooks, they are treated separately, see below
        if thing == "hooks":
            continue
        print("      {} args: {}".format(thing, args['args']))

    # check if there are any adapter specific-hooks implemented
    adapter_hook_names = feature_map.get("hooks", [])
    if adapter_hook_names:
        print("      adapter hooks: {}".format(adapter_hook_names))

    extra_features = []
    for kind in ["read", "write"]:
        if (
            f"{kind}_from_string" in feature_map
            and f"{kind}_from_file" not in feature_map
        ):
            extra_features.append(
                "{0}_from_file (calls: {0}_from_string)".format(kind)
            )

    if extra_features:
        print("    implicit supported features:")
        for feat in extra_features:
            print(f"      {feat}")


def _schemadefs_formatted(feature_map, args):
    print("    SchemaDefs:")
    for sd in feature_map.keys():
        print(f"      {sd}")
        _docs_formatted(feature_map[sd]['doc'], args, indent=8)


def _docs_formatted(docstring, arg_map, indent=4):
    long_docs = arg_map.get('long_docs')

    if long_docs:
        prefix = " " * indent + "doc (long): "
    else:
        prefix = " " * indent + "doc (short): "

    initial_indent = prefix
    subsequent_indent = " " * len(prefix)

    try:
        block = docstring.split("\n")
    except AttributeError:
        raise RuntimeError(
            "Plugin: '{}' is missing a docstring.  Make sure the doctring is "
            "assigned to the __doc__ variable name.".format(arg_map['plugname'])
        )

    fmt_block = []
    for line in block:
        line = textwrap.fill(
            line,
            initial_indent=initial_indent,
            subsequent_indent=subsequent_indent,
            width=len(subsequent_indent) + 80,
        )
        initial_indent = subsequent_indent
        fmt_block.append(line)

    if long_docs:
        text = "\n".join(fmt_block)
    else:
        text = fmt_block[0]

    print(text)


_FORMATTER = {
    "supported features": _supported_features_formatted,
    "SchemaDefs": _schemadefs_formatted,
    "doc": _docs_formatted,
}

_FIELDS_TO_SKIP = frozenset(["name"])


def _print_field(key, val, **args):
    # if attribute doesn't hit any of the user specified patterns
    if (
        not any(fnmatch.filter([key], pt) for pt in args['attribs'])
        or key in _FIELDS_TO_SKIP
    ):
        return

    if key in _FORMATTER:
        _FORMATTER[key](val, args)
        return

    print(f"    {key}: {val}")


def main():
    """  main entry point  """
    args = _parsed_args()

    plugin_types = args.plugin_types

    if 'all' in plugin_types:
        # all means the full list of built in plugins
        plugin_types = otio.plugins.manifest.OTIO_PLUGIN_TYPES

    # load all the otio plugins
    active_plugin_manifest = otio.plugins.ActiveManifest()

    # print version information to the shell
    if args.version:
        print(f"OpenTimelineIO version: {otio.__version__}")

        entry_points = otio.plugins.manifest.plugin_entry_points()
        if entry_points:
            print("Plugins from installed packages:")
            for plugin in entry_points:
                print(f"   {plugin.dist.name} {plugin.dist.version}")

    # list the loaded manifests
    print("Manifests loaded:")
    for mf in active_plugin_manifest.source_files:
        print(f"  {mf}")

    for pt in plugin_types:
        # hooks have special code (see below)
        if pt == "hooks":
            continue

        # header
        print("")
        print(f"{pt}:")

        # filter plugins by the patterns passed in on the command line
        plugin_by_type = getattr(active_plugin_manifest, pt)
        plugins = [
            p for p in plugin_by_type
            if fnmatch.filter([p.name], args.plugpattern)
        ]

        # if nothing is found of that type that matches the filter
        if not plugins:
            print("    (none found)")

        for plug in plugins:
            print(f"  {plug.name}")

            try:
                info = plug.plugin_info_map()
            except Exception as err:
                print(
                    "    ERROR: plugin {} couldn't generate its information"
                    " map: {}\n".format(
                        plug.name,
                        err
                    )
                )
                continue

            for key, val in info.items():
                _print_field(
                    key,
                    val,
                    long_docs=args.long_docs,
                    attribs=args.attribs,
                    plugname=plug.name,
                )

    # hooks aren't really plugin objects, instead they're a mapping of hook
    # to list of hookscripts that will run on that hook
    if "hooks" in plugin_types:
        print("")
        print("hooks:")
        hooknames = fnmatch.filter(
            active_plugin_manifest.hooks.keys(),
            args.plugpattern
        )
        for hookname in hooknames:
            print(f"  {hookname}")
            for hook_script in active_plugin_manifest.hooks[hookname]:
                print(f"    {hook_script}")
            if not active_plugin_manifest.hooks[hookname]:
                print("    (no hook scripts attached)")

        if not hooknames:
            print("  (none found)")


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""OTIO Python Plugin Manifest system: locates plugins to OTIO."""

from importlib import resources
import inspect
import logging
import os

# for python <= 3.9
from importlib import metadata

from .. import (
    core,
    exceptions,
)


# for tracking what kinds of plugins the manifest system supports
OTIO_PLUGIN_TYPES = [
    'adapters',
    'media_linkers',
    'schemadefs',
    'hook_scripts',
    'hooks',
    'version_manifests',
]


def plugin_entry_points():
    """Returns the list of entry points for all available OpenTimelineIO
    plugins.
    """
    try:
        entry_points = metadata.entry_points(group='opentimelineio.plugins')
    except TypeError:
        # For python <= 3.9
        entry_points = metadata.entry_points().get(
            'opentimelineio.plugins', []
        )

    return entry_points


def manifest_from_file(filepath):
    """Read the .json file at filepath into a :py:class:`Manifest` object."""

    result = core.deserialize_json_from_file(filepath)
    absfilepath = os.path.abspath(filepath)
    result.source_files.append(absfilepath)
    result._update_plugin_source(absfilepath)
    return result


def manifest_from_string(input_string):
    """Deserialize the json string into a manifest object."""

    result = core.deserialize_json_from_string(input_string)

    # try and get the caller's name
    name = "unknown"
    stack = inspect.stack()
    if len(stack) > 1 and len(stack[1]) > 3:
        #                     filename     function name
        name = f"{stack[1][1]}:{stack[1][3]}"

    # set the value in the manifest
    src_string = f"call to manifest_from_string() in {name}"
    result.source_files.append(src_string)
    result._update_plugin_source(src_string)

    return result


@core.register_type
class Manifest(core.SerializableObject):
    """Defines an OTIO plugin Manifest.

    This is considered an internal OTIO implementation detail.

    A manifest tracks a collection of plugins and enables finding them by name
    or other features (in the case of adapters, what file suffixes they
    advertise support for).

    For more information, consult the documentation.
    """
    _serializable_label = "PluginManifest.1"

    def __init__(self):
        core.SerializableObject.__init__(self)
        self.adapters = []
        self.schemadefs = []
        self.media_linkers = []
        self.source_files = []

        # hook system stuff
        self.hooks = {}
        self.hook_scripts = []

        self.version_manifests = {}

    adapters = core.serializable_field(
        "adapters",
        type([]),
        "Adapters this manifest describes."
    )
    schemadefs = core.serializable_field(
        "schemadefs",
        type([]),
        "Schemadefs this manifest describes."
    )
    media_linkers = core.serializable_field(
        "media_linkers",
        type([]),
        "Media Linkers this manifest describes."
    )
    hooks = core.serializable_field(
        "hooks",
        type({}),
        "Hooks that hooks scripts can be attached to."
    )
    hook_scripts = core.serializable_field(
        "hook_scripts",
        type([]),
        "Scripts that can be attached to hooks."
    )
    version_manifests = core.serializable_field(
        "version_manifests",
        type({}),
        "Sets of versions to downgrade schemas to."
    )

    def extend(self, another_manifest):
        """
        Aggregate another manifest's plugins into this one.

        During startup, OTIO will deserialize the individual manifest JSON files
        and use this function to concatenate them together.
        """
        if not another_manifest:
            return

        self.adapters.extend(another_manifest.adapters)
        self.schemadefs.extend(another_manifest.schemadefs)
        self.media_linkers.extend(another_manifest.media_linkers)
        self.hook_scripts.extend(another_manifest.hook_scripts)

        for family, label_map in another_manifest.version_manifests.items():
            # because self.version_manifests is an AnyDictionary instead of a
            # vanilla python dictionary, it does not support the .set_default()
            # method.
            if family not in self.version_manifests:
                self.version_manifests[family] = {}
            self.version_manifests[family].update(label_map)

        for trigger_name, hooks in another_manifest.hooks.items():
            # because self.hooks is an AnyDictionary instead of a vanilla
            # python dictionary, it does not support the .set_default() method.
            if trigger_name not in self.hooks:
                self.hooks[trigger_name] = []
            self.hooks[trigger_name].extend(hooks)

        self.source_files.extend(another_manifest.source_files)

    def _update_plugin_source(self, path):
        """Set the source file path for the manifest."""

        for thing in (
            self.adapters
            + self.schemadefs
            + self.media_linkers
            + self.hook_scripts
        ):
            thing._json_path = path

    def from_filepath(self, suffix):
        """Return the adapter object associated with a given file suffix."""

        for adapter in self.adapters:
            if suffix.lower() in adapter.suffixes:
                return adapter
        raise exceptions.NoKnownAdapterForExtensionError(suffix)

    def adapter_module_from_suffix(self, suffix):
        """Return the adapter module associated with a given file suffix."""

        adp = self.from_filepath(suffix)
        return adp.module()

    # @TODO: (breaking change) this should search all plugins by default instead
    #        of just adapters
    def from_name(self, name, kind_list="adapters"):
        """Return the plugin object associated with a given plugin name."""

        for thing in getattr(self, kind_list):
            if name == thing.name:
                return thing

        raise exceptions.NotSupportedError(
            "Could not find plugin: '{}' in kind_list: '{}'."
            " options: {}".format(
                name,
                kind_list,
                getattr(self, kind_list)
            )
        )

    def adapter_module_from_name(self, name):
        """Return the adapter module associated with a given adapter name."""

        adp = self.from_name(name)
        return adp.module()

    def schemadef_module_from_name(self, name):
        """Return the schemadef module associated with a given schemadef name."""

        adp = self.from_name(name, kind_list="schemadefs")
        return adp.module()


_MANIFEST = None


def load_manifest():
    """Walk the plugin manifest discovery systems and accumulate manifests.

    The order of loading (and precedence) is:

       1. Manifests specified via the :term:`OTIO_PLUGIN_MANIFEST_PATH` variable
       2. Entrypoint based plugin manifests
       3. Builtin plugin manifest
    """

    result = Manifest()

    # Read plugin manifests defined on the $OTIO_PLUGIN_MANIFEST_PATH
    # environment variable.  This variable is an os.pathsep separated list of
    # file paths to manifest json files. Can be set to "" to indicate no
    # local custom manifest path.
    _local_manifest_path = os.environ.get("OTIO_PLUGIN_MANIFEST_PATH", None)
    if _local_manifest_path:
        for src_json_path in _local_manifest_path.split(os.pathsep):
            json_path = os.path.abspath(src_json_path)
            if (
                not os.path.exists(json_path)
                # the manifest has already been loaded
                or json_path in result.source_files
            ):
                # XXX: In case error reporting is requested
                # print(
                #     "Warning: OpenTimelineIO cannot access path '{}' from "
                #     "$OTIO_PLUGIN_MANIFEST_PATH".format(json_path)
                # )
                continue

            result.extend(manifest_from_file(json_path))

    if not os.environ.get("OTIO_DISABLE_ENTRYPOINTS_PLUGINS"):
        entry_points = plugin_entry_points()

        for plugin in entry_points:
            plugin_name = plugin.name
            try:
                plugin_entry_point = plugin.load()
                try:
                    plugin_manifest = plugin_entry_point.plugin_manifest()

                    # this ignores what the plugin_manifest() function might
                    # put into source_files in favor of using the path to the
                    # python package as the unique identifier

                    manifest_path = os.path.abspath(
                        plugin_entry_point.__file__
                    )

                    if manifest_path in result.source_files:
                        continue

                    plugin_manifest.source_files = [manifest_path]
                    plugin_manifest._update_plugin_source(manifest_path)

                except AttributeError:
                    name = plugin_entry_point.__name__

                    filepath = resources.files(name) / "plugin_manifest.json"

                    if filepath.as_posix() in result.source_files:
                        continue

                    plugin_manifest = core.deserialize_json_from_string(
                        filepath.read_text()
                    )
                    plugin_manifest._update_plugin_source(filepath.as_posix())
                    plugin_manifest.source_files.append(filepath.as_posix())

            except Exception as e:
                logging.exception(
                    f"could not load plugin: {plugin_name}.  Exception is: {e}"
                )
                continue

            result.extend(plugin_manifest)
    else:
        logging.debug(
            "OTIO_DISABLE_ENTRYPOINTS_PLUGINS is set. "
            "Entry points plugings have been disabled."
        )

    # the builtin plugin manifest
    builtin_manifest_path = (
        resources.files("opentimelineio.adapters")
        / "builtin_adapters.plugin_manifest.json"
    ).as_posix()

    if os.path.abspath(builtin_manifest_path) not in result.source_files:
        plugin_manifest = manifest_from_file(builtin_manifest_path)
        result.extend(plugin_manifest)

    # force the schemadefs to load and add to schemadef module namespace
    for s in result.schemadefs:
        s.module()

    return result


def ActiveManifest(force_reload=False):
    """Return the fully resolved plugin manifest."""

    global _MANIFEST
    if not _MANIFEST or force_reload:
        _MANIFEST = load_manifest()

    return _MANIFEST



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

from .. core._core_utils import add_method
from .. import _otio


@add_method(_otio.Timeline)
def __str__(self):
    return f'Timeline("{str(self.name)}", {str(self.tracks)})'


@add_method(_otio.Timeline)
def __repr__(self):
    return (
        "otio.schema.Timeline(name={}, tracks={})".format(
            repr(self.name),
            repr(self.tracks)
        )
    )



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""An editorial interchange format and library.

see: http://opentimeline.io

.. moduleauthor:: Contributors to the OpenTimelineIO project <otio-discussion@lists.aswf.io>
"""

# flake8: noqa

# in dependency hierarchy
from . import (
    opentime,
    exceptions,
    core,
    schema,
    schemadef,
    plugins,
    media_linker,
    adapters,
    hooks,
    algorithms,
    url_utils,
    versioning,
)

__version__ = "0.18.1"
__author__ = "Contributors to the OpenTimelineIO project"
__author_email__ = "otio-discussion@lists.aswf.io"
__license__ = "Apache 2.0 License"



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Core implementation details and wrappers around the C++ library"""

from .. _otio import ( # noqa
    # errors
    CannotComputeAvailableRangeError,

    # classes
    Color,
    Composable,
    Composition,
    Item,
    MediaReference,
    SerializableObject,
    SerializableObjectWithMetadata,
    Track,

    # functions
    deserialize_json_from_file,
    deserialize_json_from_string,
    flatten_stack,
    install_external_keepalive_monitor,
    instance_from_schema,
    register_serializable_object_type,
    register_upgrade_function,
    register_downgrade_function,
    set_type_record,
    _serialize_json_to_string,
    _serialize_json_to_file,
    type_version_map,
    release_to_schema_version_map,
)

from . _core_utils import ( # noqa
    add_method,
    _value_to_any,
    _add_mutable_mapping_methods,
    _add_mutable_sequence_methods,
)
from . import ( # noqa
    mediaReference,
    color,
    composition,
    composable,
    item,
)

__all__ = [
    'Color',
    'Composable',
    'Composition',
    'Item',
    'MediaReference',
    'SerializableObject',
    'SerializableObjectWithMetadata',
    'Track',
    'deserialize_json_from_file',
    'deserialize_json_from_string',
    'flatten_stack',
    'install_external_keepalive_monitor',
    'instance_from_schema',
    'set_type_record',
    'add_method',
    'upgrade_function_for',
    'downgrade_function_from',
    'serializable_field',
    'deprecated_field',
    'serialize_json_to_string',
    'serialize_json_to_file',
    'register_type',
    'type_version_map',
    'release_to_schema_version_map',
]


def serialize_json_to_string(root, schema_version_targets=None, indent=4):
    """Serialize root to a json string.  Optionally downgrade resulting schemas
    to schema_version_targets.

    :param SerializableObject root: root object to serialize
    :param dict[str, int] schema_version_targets: optional dictionary mapping
                                                  schema name to desired schema
                                                  version, for downgrading the
                                                  result to be compatible with
                                                  older versions of
                                                  OpenTimelineIO.
    :param int indent: number of spaces for each json indentation level. Use -1
                       for no indentation or newlines.

    :returns: resulting json string
    :rtype: str
    """
    return _serialize_json_to_string(
        _value_to_any(root),
        schema_version_targets or {},
        indent
    )


def serialize_json_to_file(
        root,
        filename,
        schema_version_targets=None,
        indent=4
):
    """Serialize root to a json file.  Optionally downgrade resulting schemas
    to schema_version_targets.

    :param SerializableObject root: root object to serialize
    :param dict[str, int] schema_version_targets: optional dictionary mapping
                                                  schema name to desired schema
                                                  version, for downgrading the
                                                  result to be compatible with
                                                  older versions of
                                                  OpenTimelineIO.
    :param int indent: number of spaces for each json indentation level. Use -1
                       for no indentation or newlines.

    :returns: true for success, false for failure
    :rtype: bool
    """
    return _serialize_json_to_file(
        _value_to_any(root),
        filename,
        schema_version_targets or {},
        indent
    )


def register_type(classobj, schemaname=None):
    """Decorator for registering a SerializableObject type

    Example:

    .. code-block:: python

        @otio.core.register_type
        class SimpleClass(otio.core.SerializableObject):
          serializable_label = "SimpleClass.2"
          ...

    :param typing.Type[SerializableObject] cls: class to register
    :param str schemaname: Schema name (default: parse from serializable_label)
    """
    label = classobj._serializable_label
    if schemaname is None:
        schema_name, schema_version = label.split(".", 2)
    else:
        schema_name, schema_version = schemaname, 1

    register_serializable_object_type(classobj, schema_name, int(schema_version))

    orig_init = classobj.__init__

    def __init__(self, *args, **kwargs):
        orig_init(self, *args, **kwargs)
        set_type_record(self, schema_name)

    classobj.__init__ = __init__
    return classobj


def upgrade_function_for(cls, version_to_upgrade_to):
    """
    Decorator for identifying schema class upgrade functions.

    Example:

    .. code-block:: python

        @upgrade_function_for(MyClass, 5)
        def upgrade_to_version_five(data):
            pass

    This will get called to upgrade a schema of MyClass to version 5. MyClass
    must be a class deriving from :class:`~SerializableObject`.

    The upgrade function should take a single argument - the dictionary to
    upgrade, and return a dictionary with the fields upgraded.

    Remember that you don't need to provide an upgrade function for upgrades
    that add or remove fields, only for schema versions that change the field
    names.

    :param typing.Type[SerializableObject] cls: class to upgrade
    :param int version_to_upgrade_to: the version to upgrade to
    """

    def decorator_func(func):
        """ Decorator for marking upgrade functions """
        def wrapped_update(data):
            modified = func(data)
            data.clear()
            data.update(modified)

        register_upgrade_function(cls._serializable_label.split(".")[0],
                                  version_to_upgrade_to, wrapped_update)
        return func

    return decorator_func


def downgrade_function_from(cls, version_to_downgrade_from):
    """
    Decorator for identifying schema class downgrade functions.

    Example:

    .. code-block:: python

        @downgrade_function_from(MyClass, 5)
        def downgrade_from_five_to_four(data):
            return {"old_attr": data["new_attr"]}

    This will get called to downgrade a schema of MyClass from version 5 to
    version 4. MyClass must be a class deriving from
    :class:`~SerializableObject`.

    The downgrade function should take a single argument - the dictionary to
    downgrade, and return a dictionary with the fields downgraded.

    :param typing.Type[SerializableObject] cls: class to downgrade
    :param int version_to_downgrade_from: the function downgrading from this
                                          version to (version - 1)
    """

    def decorator_func(func):
        """ Decorator for marking downgrade functions """
        def wrapped_update(data):
            modified = func(data)
            data.clear()
            data.update(modified)

        register_downgrade_function(
            cls._serializable_label.split(".")[0],
            version_to_downgrade_from,
            wrapped_update
        )
        return func

    return decorator_func


def serializable_field(name, required_type=None, doc=None, default_value=None):
    """
    Convenience function for adding attributes to child classes of
    :class:`~SerializableObject` in such a way that they will be serialized/deserialized
    automatically.

    Use it like this:

    .. code-block:: python

        @core.register_type
        class Foo(SerializableObject):
            bar = serializable_field("bar", required_type=int, doc="example")

    This would indicate that class "foo" has a serializable field "bar".  So:

    .. code-block:: python

        f = foo()
        f.bar = "stuff"

        # serialize & deserialize
        otio_json = otio.adapters.from_name("otio")
        f2 = otio_json.read_from_string(otio_json.write_to_string(f))

        # fields should be equal
        f.bar == f2.bar

    Additionally, the "doc" field will become the documentation for the
    property.

    :param str name: name of the field to add
    :param type required_type: type required for the field
    :param str doc: field documentation
    :param Any default_value: default value to return if no field value is set yet

    :returns: property object
    :rtype: :py:class:`property`
    """

    def getter(self):
        return self._dynamic_fields.get(name, default_value)

    def setter(self, val):
        # always allow None values regardless of value of required_type
        if required_type is not None and val is not None:
            if not isinstance(val, required_type):
                raise TypeError(
                    "attribute '{}' must be an instance of '{}', not: {}".format(
                        name,
                        required_type,
                        type(val)
                    )
                )

        self._dynamic_fields[name] = val

    return property(getter, setter, doc=doc)


def deprecated_field():
    """For marking attributes on a SerializableObject deprecated."""

    def getter(self):
        raise DeprecationWarning

    def setter(self, val):
        raise DeprecationWarning

    return property(getter, setter, doc="Deprecated field, do not use.")



#!/usr/bin/env python
#
# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

import argparse
import sys
import copy

import opentimelineio as otio

# on some python interpreters, pkg_resources is not available
try:
    import pkg_resources
except ImportError:
    pkg_resources = None

__doc__ = """ Python wrapper around OTIO to convert timeline files between \
formats.

Available adapters: {}
""".format(otio.adapters.available_adapter_names())


def _parsed_args():
    """ parse commandline arguments with argparse """

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        '-i',
        '--input',
        type=str,
        required=False,
        help='path to input file',
    )
    parser.add_argument(
        '-o',
        '--output',
        type=str,
        required=False,
        help='path to output file',
    )
    parser.add_argument(
        '-I',
        '--input-adapter',
        type=str,
        default=None,
        help="Explicitly use this adapter for reading the input file",
    )
    parser.add_argument(
        '-O',
        '--output-adapter',
        type=str,
        default=None,
        help="Explicitly use this adapter for writing the output file",
    )
    parser.add_argument(
        '-T',
        '--tracks',
        type=str,
        default=None,
        help="Pick one or more tracks, by 0-based index, separated by commas.",
    )
    parser.add_argument(
        '-m',
        '--media-linker',
        type=str,
        default="Default",
        help=(
            "Specify a media linker.  'Default' means use the "
            "$OTIO_DEFAULT_MEDIA_LINKER if set, 'None' or '' means explicitly "
            "disable the linker, and anything else is interpreted as the name"
            " of the media linker to use."
        )
    )
    parser.add_argument(
        '-H',
        '--hook-function-arg',
        type=str,
        default=[],
        action='append',
        help='Extra arguments to be passed to the hook functions in the form of '
        'key=value. Values are strings, numbers or Python literals: True, '
        'False, etc. Can be used multiple times: -H burrito="bar" -H taco=12.'
    )
    parser.add_argument(
        '-M',
        '--media-linker-arg',
        type=str,
        default=[],
        action='append',
        help='Extra arguments to be passed to the media linker in the form of '
        'key=value. Values are strings, numbers or Python literals: True, '
        'False, etc. Can be used multiple times: -M burrito="bar" -M taco=12.'
    )
    parser.add_argument(
        '-a',
        '--adapter-arg',
        type=str,
        default=[],
        action='append',
        help='Extra arguments to be passed to input adapter in the form of '
        'key=value. Values are strings, numbers or Python literals: True, '
        'False, etc. Can be used multiple times: -a burrito="bar" -a taco=12.'
    )
    parser.add_argument(
        '-A',
        '--output-adapter-arg',
        type=str,
        default=[],
        action='append',
        help='Extra arguments to be passed to output adapter in the form of '
        'key=value. Values are strings, numbers or Python literals: True, '
        'False, etc. Can be used multiple times: -A burrito="bar" -A taco=12.'
    )
    parser.add_argument(
        '--version',
        default=False,
        action="store_true",
        help=(
            "Print the otio and pkg_resource installed plugin version "
            "information to the commandline and then exit."
        ),
    )

    trim_args = parser.add_argument_group(
        title="Trim Arguments",
        description="Arguments that allow you to trim the OTIO file."
    )
    trim_args.add_argument(
        '--begin',
        type=str,
        default=None,
        help=(
            "Trim out everything in the timeline before this time, in the "
            "global time frame of the timeline.  Argument should be in the form"
            ' "VALUE,RATE", eg: --begin "10,24".  Requires --end argument.'
        ),
    )
    trim_args.add_argument(
        '--end',
        type=str,
        default=None,
        help=(
            "Trim out everything in the timeline after this time, in the "
            "global time frame of the timeline.  Argument should be in the form"
            ' "VALUE,RATE", eg: --begin "10,24".  Requires --begin argument.'
        ),
    )

    result = parser.parse_args()

    # print version information to the shell
    if result.version:
        print(f"OpenTimelineIO version: {otio.__version__}")

        entry_points = otio.plugins.manifest.plugin_entry_points()
        if entry_points:
            print("Plugins from installed packages:")
            for plugin in entry_points:
                print(f"   {plugin.dist.name} {plugin.dist.version}")
        parser.exit()

    if not result.input:
        parser.error("-i/--input is a required argument")
    if not result.output:
        parser.error("-o/--output is a required argument")

    if result.begin is not None and result.end is None:
        parser.error("--begin requires --end.")
    if result.end is not None and result.begin is None:
        parser.error("--end requires --begin.")

    if result.begin is not None:
        try:
            value, rate = result.begin.split(",")
            result.begin = otio.opentime.RationalTime(float(value), float(rate))
        except ValueError:
            parser.error(
                "--begin argument needs to be of the form: VALUE,RATE where "
                "VALUE is the (float) time value of the resulting RationalTime "
                "and RATE is the (float) time rate of the resulting RationalTime,"
                " not '{}'".format(result.begin)
            )

    if result.end is not None:
        try:
            value, rate = result.end.split(",")
            result.end = otio.opentime.RationalTime(float(value), float(rate))
        except ValueError:
            parser.error(
                "--end argument needs to be of the form: VALUE,RATE where "
                "VALUE is the (float) time value of the resulting RationalTime "
                "and RATE is the (float) time rate of the resulting RationalTime,"
                " not '{}'".format(result.begin)
            )

    return result


def main():
    """Parse arguments and convert the files."""

    args = _parsed_args()

    in_adapter = args.input_adapter
    if in_adapter is None:
        in_adapter = otio.adapters.from_filepath(args.input).name

    out_adapter = args.output_adapter
    if out_adapter is None:
        out_adapter = otio.adapters.from_filepath(args.output).name

    media_linker_name = otio.console.console_utils.media_linker_name(
        args.media_linker
    )

    try:
        read_adapter_arg_map = otio.console.console_utils.arg_list_to_map(
            args.adapter_arg,
            "input adapter"
        )
        hooks_args = otio.console.console_utils.arg_list_to_map(
            args.hook_function_arg,
            "hook function"
        )
        ml_args = otio.console.console_utils.arg_list_to_map(
            args.media_linker_arg,
            "media linker"
        )
    except ValueError as exc:
        sys.stderr.write("\n" + str(exc) + "\n")
        sys.exit(1)

    result_tl = otio.adapters.read_from_file(
        args.input,
        in_adapter,
        hook_function_argument_map=hooks_args,
        media_linker_name=media_linker_name,
        media_linker_argument_map=ml_args,
        **read_adapter_arg_map
    )

    if args.tracks:
        result_tracks = copy.deepcopy(otio.schema.Stack())
        del result_tracks[:]
        for track in args.tracks.split(","):
            tr = result_tl.tracks[int(track)]
            del result_tl.tracks[int(track)]
            print(f"track {track} is of kind: '{tr.kind}'")
            result_tracks.append(tr)
        result_tl.tracks = result_tracks

    # handle trim arguments
    if args.begin is not None and args.end is not None:
        result_tl = otio.algorithms.timeline_trimmed_to_range(
            result_tl,
            otio.opentime.range_from_start_end_time(args.begin, args.end)
        )

    try:
        write_adapter_arg_map = otio.console.console_utils.arg_list_to_map(
            args.output_adapter_arg,
            "output adapter"
        )
    except ValueError as exc:
        sys.stderr.write("\n" + str(exc) + "\n")
        sys.exit(1)

    otio.adapters.write_to_file(
        result_tl,
        args.output,
        out_adapter,
        hook_function_argument_map=hooks_args,
        **write_adapter_arg_map
    )


if __name__ == '__main__':
    try:
        main()
    except otio.exceptions.OTIOError as err:
        sys.stderr.write("ERROR: " + str(err) + "\n")
        sys.exit(1)



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

__doc__ = """ Algorithms for stack objects. """

from .. import (
    schema,
    opentime,
    _otio,
)


def top_clip_at_time(in_stack, t):
    """Return the topmost visible child that overlaps with time ``t``.

    Example::

        tr1: G1, A, G2
        tr2: [B------]
        G1, and G2 are gaps, A and B are clips.

    If ``t`` is within ``A``, ``A`` will be returned. If ``t`` is within ``G1`` or
    ``G2``, ``B`` will be returned.

    :param Stack in_stack: Stack
    :param RationalTime t: Time
    :returns: Top clip
    :rtype: Clip
    """

    # ensure that it only runs on stacks
    if not isinstance(in_stack, schema.Stack):
        raise ValueError(
            "Argument in_stack must be of type otio.schema.Stack, "
            "not: '{}'".format(
                type(in_stack)
            )
        )

    # build a range to use the `find_clips`method.
    search_range = opentime.TimeRange(
        start_time=t,
        # 0 duration so we are just sampling a point in time.
        # XXX Should this duration be equal to the length of one sample?
        #     opentime.RationalTime(1, rate)?
        duration=opentime.RationalTime(0, t.rate)
    )

    # walk through the children of the stack in reverse order.
    for track in reversed(in_stack):
        valid_results = []
        if hasattr(track, "find_clips"):
            valid_results = list(
                c for c in track.find_clips(search_range, shallow_search=True)
                if c.visible()
            )

        # XXX doesn't handle nested tracks/stacks at the moment

        for result in valid_results:
            return result

    return None


flatten_stack = _otio.flatten_stack



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

from .. core._core_utils import add_method
from .. import _otio


@add_method(_otio.ExternalReference)
def __str__(self):
    return f'ExternalReference("{str(self.target_url)}")'


@add_method(_otio.ExternalReference)
def __repr__(self):
    return 'otio.schema.ExternalReference(target_url={})'.format(
        repr(str(self.target_url))
    )



#!/usr/bin/env python
#
# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Print the contents of an OTIO file to stdout."""

import argparse
import sys

import opentimelineio as otio


def _parsed_args():
    """ parse commandline arguments with argparse """

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        'filepath',
        type=str,
        nargs='+',
        help='files to print the contents of'
    )
    parser.add_argument(
        '-a',
        '--adapter-arg',
        type=str,
        default=[],
        action='append',
        help='Extra arguments to be passed to input adapter in the form of '
        'key=value. Values are strings, numbers or Python literals: True, '
        'False, etc. Can be used multiple times: -a burrito="bar" -a taco=12.'
    )
    parser.add_argument(
        '-m',
        '--media-linker',
        type=str,
        default="Default",
        help=(
            "Specify a media linker.  'Default' means use the "
            "$OTIO_DEFAULT_MEDIA_LINKER if set, 'None' or '' means explicitly "
            "disable the linker, and anything else is interpreted as the name"
            " of the media linker to use."
        )
    )
    parser.add_argument(
        '-H',
        '--hook-function-arg',
        type=str,
        default=[],
        action='append',
        help='Extra arguments to be passed to the hook functions in the form of '
        'key=value. Values are strings, numbers or Python literals: True, '
        'False, etc. Can be used multiple times: -H burrito="bar" -H taco=12.'
    )
    parser.add_argument(
        '-M',
        '--media-linker-arg',
        type=str,
        default=[],
        action='append',
        help='Extra arguments to be passed to the media linker in the form of '
        'key=value. Values are strings, numbers or Python literals: True, '
        'False, etc. Can be used multiple times: -M burrito="bar" -M taco=12.'
    )

    return parser.parse_args()


def _otio_compatible_file_to_json_string(
        fpath,
        media_linker_name,
        hooks_args,
        media_linker_argument_map,
        adapter_argument_map
):
    """Read the file at fpath with the default otio adapter and return the json
    as a string.
    """

    adapter = otio.adapters.from_name("otio_json")
    return adapter.write_to_string(
        otio.adapters.read_from_file(
            fpath,
            hook_function_argument_map=hooks_args,
            media_linker_name=media_linker_name,
            media_linker_argument_map=media_linker_argument_map,
            **adapter_argument_map
        )
    )


def main():
    """Parse arguments and call _otio_compatible_file_to_json_string."""

    args = _parsed_args()

    media_linker_name = otio.console.console_utils.media_linker_name(
        args.media_linker
    )

    try:
        read_adapter_arg_map = otio.console.console_utils.arg_list_to_map(
            args.adapter_arg,
            "adapter"
        )
        hooks_args = otio.console.console_utils.arg_list_to_map(
            args.hook_function_arg,
            "hook function"
        )
        media_linker_argument_map = otio.console.console_utils.arg_list_to_map(
            args.media_linker_arg,
            "media linker"
        )
    except ValueError as exc:
        sys.stderr.write("\n" + str(exc) + "\n")
        sys.exit(1)

    for fpath in args.filepath:
        print(
            _otio_compatible_file_to_json_string(
                fpath,
                media_linker_name,
                hooks_args,
                media_linker_argument_map,
                read_adapter_arg_map
            )
        )


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

from . _core_utils import add_method
from .. import _otio


@add_method(_otio.Item)
def __str__(self):
    return "{}({}, {}, {}, {}, {}, {})".format(
        self.__class__.__name__,
        self.name,
        str(self.source_range),
        str(self.effects),
        str(self.markers),
        str(self.enabled),
        str(self.metadata)
    )


@add_method(_otio.Item)
def __repr__(self):
    return (
        "otio.{}.{}("
        "name={}, "
        "source_range={}, "
        "effects={}, "
        "markers={}, "
        "enabled={}, "
        "metadata={}"
        ")".format(
            "core" if self.__class__ is _otio.Item else "schema",
            self.__class__.__name__,
            repr(self.name),
            repr(self.source_range),
            repr(self.effects),
            repr(self.markers),
            repr(self.enabled),
            repr(self.metadata)
        )
    )



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Algorithms for timeline objects."""

import copy

from . import (
    track_algo
)


def timeline_trimmed_to_range(in_timeline, trim_range):
    """
    Returns a new timeline that is a copy of the in_timeline, but with items
    outside the trim_range removed and items on the ends trimmed to the
    trim_range.

    .. note:: the timeline is never expanded, only shortened.

    Please note that you could do nearly the same thing non-destructively by
    just setting the :py:class:`.Track`\'s source_range but sometimes you want to
    really cut away the stuff outside and that's what this function is meant for.

    :param Timeline in_timeline: Timeline to trim
    :param TimeRange trim_range:
    :returns: New trimmed timeline
    :rtype: Timeline
    """
    new_timeline = copy.deepcopy(in_timeline)

    for track_num, child_track in enumerate(in_timeline.tracks):
        # @TODO: put the trim_range into the space of the tracks
        # new_range = new_timeline.tracks.transformed_time_range(
        #     trim_range,
        #     child_track
        # )

        # trim the track and assign it to the new stack.
        new_timeline.tracks[track_num] = track_algo.track_trimmed_to_range(
            child_track,
            trim_range
        )

    return new_timeline



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Expose the adapter interface to developers.

To read from an existing representation, use the read_from_string and
read_from_file functions.  To query the list of adapters, use the
available_adapter_names function.

The otio_json adapter is provided as a the canonical, lossless, serialization
of the in-memory otio schema.  Other adapters are to varying degrees lossy.
For more information, consult the documentation in the individual adapter
modules.
"""

import os
import itertools
import pathlib

from .. import (
    exceptions,
    plugins,
    media_linker
)

from .adapter import Adapter  # noqa

# OTIO Json, OTIOZ and OTIOD adapters are always available
from . import (  # noqa: F401
    otio_json,  # core JSON adapter
    file_bundle_utils,  # utilities for working with OTIO file bundles
)

__all__ = [
    'Adapter',
    'otio_json',
    'file_bundle_utils',
    'suffixes_with_defined_adapters',
    'available_adapter_names',
    'from_filepath',
    'from_name',
    'read_from_file',
    'read_from_string',
    'write_to_file',
    'write_to_string'
]


def suffixes_with_defined_adapters(read=False, write=False):
    """Return a set of all the suffixes that have adapters defined for them."""

    if not read and not write:
        read = True
        write = True

    positive_adapters = []
    for adp in plugins.ActiveManifest().adapters:
        if read and adp.has_feature("read"):
            positive_adapters.append(adp)
            continue

        if write and adp.has_feature("write"):
            positive_adapters.append(adp)

    return set(
        itertools.chain.from_iterable(
            adp.suffixes for adp in positive_adapters
        )
    )


def available_adapter_names():
    """Return a string list of the available adapters."""

    return [str(adp.name) for adp in plugins.ActiveManifest().adapters]


def _from_filepath_or_name(filepath, adapter_name):
    if adapter_name is not None:
        return plugins.ActiveManifest().from_name(adapter_name)
    else:
        return from_filepath(filepath)


def from_filepath(filepath):
    """Guess the adapter object to use for a given filepath.

    For example, ``foo.otio`` returns the ``otio_json`` adapter.
    """

    outext = os.path.splitext(filepath)[1][1:]

    try:
        return plugins.ActiveManifest().from_filepath(outext)
    except exceptions.NoKnownAdapterForExtensionError:
        raise exceptions.NoKnownAdapterForExtensionError(
            "No adapter for suffix '{}' on file '{}'".format(
                outext,
                filepath
            )
        )


def from_name(name):
    """Fetch the adapter object by the name of the adapter directly."""

    try:
        return plugins.ActiveManifest().from_name(name)
    except exceptions.NotSupportedError:
        raise exceptions.NotSupportedError(
            "adapter not supported: {}, available: {}".format(
                name,
                available_adapter_names()
            )
        )


def read_from_file(
    filepath,
    adapter_name=None,
    media_linker_name=media_linker.MediaLinkingPolicy.ForceDefaultLinker,
    media_linker_argument_map=None,
    **adapter_argument_map
):
    """Read filepath using adapter_name.

    If adapter_name is None, try and infer the adapter name from the filepath.

    .. code-block:: python
       :caption: Example

        timeline = read_from_file("example_trailer.otio")
        timeline = read_from_file("file_with_no_extension", "cmx_3600")
    """

    # convert pathlib Path objects to simple string
    string_filepath = filepath
    if isinstance(string_filepath, pathlib.PurePath):
        string_filepath = os.fspath(filepath)

    adapter = _from_filepath_or_name(string_filepath, adapter_name)

    return adapter.read_from_file(
        filepath=string_filepath,
        media_linker_name=media_linker_name,
        media_linker_argument_map=media_linker_argument_map,
        **adapter_argument_map
    )


def read_from_string(
    input_str,
    adapter_name='otio_json',
    media_linker_name=media_linker.MediaLinkingPolicy.ForceDefaultLinker,
    media_linker_argument_map=None,
    **adapter_argument_map
):
    """Read a timeline from input_str using adapter_name.

    This is useful if you obtain a timeline from someplace other than the
    filesystem.

    .. code-block:: python
       :caption: Example

        raw_text = urlopen(my_url).read()
        timeline = read_from_string(raw_text, "otio_json")
    """

    adapter = plugins.ActiveManifest().from_name(adapter_name)
    return adapter.read_from_string(
        input_str=input_str,
        media_linker_name=media_linker_name,
        media_linker_argument_map=media_linker_argument_map,
        **adapter_argument_map
    )


def write_to_file(
    input_otio,
    filepath,
    adapter_name=None,
    **adapter_argument_map
):
    """Write input_otio to filepath using adapter_name.

    If adapter_name is None, infer the adapter_name to use based on the
    filepath.

    .. code-block:: python
       :caption: Example

        otio.adapters.write_to_file(my_timeline, "output.otio")
    """

    adapter = _from_filepath_or_name(filepath, adapter_name)

    # convert pathlib Path objects to simple string
    string_filepath = filepath
    if isinstance(string_filepath, pathlib.PurePath):
        string_filepath = os.fspath(filepath)

    return adapter.write_to_file(
        input_otio=input_otio,
        filepath=string_filepath,
        **adapter_argument_map
    )


def write_to_string(
    input_otio,
    adapter_name='otio_json',
    **adapter_argument_map
):
    """Return input_otio written to a string using adapter_name.

    .. code-block:: python
       :caption: Example

        raw_text = otio.adapters.write_to_string(my_timeline, "otio_json")
    """

    adapter = plugins.ActiveManifest().from_name(adapter_name)
    return adapter.write_to_string(
        input_otio=input_otio,
        **adapter_argument_map
    )



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

from . _core_utils import add_method
from .. import _otio


@add_method(_otio.MediaReference)
def __str__(self):
    return "{}({}, {}, {}, {})".format(
        self.__class__.__name__,
        repr(self.name),
        repr(self.available_range),
        repr(self.available_image_bounds),
        repr(self.metadata)
    )


@add_method(_otio.MediaReference)
def __repr__(self):
    return (
        "otio.{}.{}("
        "name={},"
        " available_range={},"
        " available_image_bounds={},"
        " metadata={}"
        ")"
    ).format(
        "core" if self.__class__ is _otio.MediaReference else "schema",
        self.__class__.__name__,
        repr(self.name),
        repr(self.available_range),
        repr(self.available_image_bounds),
        repr(self.metadata)
    )



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Algorithms for filtering OTIO files.  """

import copy

from .. import (
    schema
)


def _is_in(thing, container):
    return any(thing is item for item in container)


def _isinstance_in(child, typelist):
    return any(isinstance(child, t) for t in typelist)


def filtered_composition(
    root,
    unary_filter_fn,
    types_to_prune=None,
):
    """
    Filter a deep copy of root (and children) with ``unary_filter_fn``.

    The ``unary_filter_fn`` must have this signature:

    .. py:function:: func(item: typing.Any) -> list[typing.Any]
        :noindex:


    1. Make a deep copy of root
    2. Starting with root, perform a depth first traversal
    3. For each item (including root):

       a. If ``types_to_prune`` is not None and item is an instance of a type in
          ``types_to_prune``, prune it from the copy, continue.
       b. Otherwise, pass the copy to ``unary_filter_fn``.  If ``unary_filter_fn``:

          I.   Returns an object: add it to the copy, replacing original
          II.  Returns a tuple: insert it into the list, replacing original
          III. Returns None: prune it
    4. If an item is pruned, do not traverse its children
    5. Return the new deep copy.

    Example 1 (filter)::

        If your unary function is:
            def fn(thing):
                if thing.name == B:
                    return thing' # some transformation of B
                else:
                    return thing

        If you have a track: [A,B,C]

        filtered_composition(track, fn) => [A,B',C]

    Example 2 (prune)::

        If your unary function is:
            def fn(thing):
                if thing.name == B:
                    return None
                else:
                    return thing

        filtered_composition(track, fn) => [A,C]

    Example 3 (expand)::

        If your unary function is:
            def fn(thing):
                if thing.name == B:
                    return tuple(B_1,B_2,B_3)
                else:
                    return thing

        filtered_composition(track, fn) => [A,B_1,B_2,B_3,C]

    Example 4 (prune gaps)::

        track :: [Gap, A, Gap]
            filtered_composition(
                track, lambda _:_, types_to_prune=(otio.schema.Gap,)) => [A]

    :param SerializableObjectWithMetadata root: Object to filter on
    :param unary_filter_fn: Filter function
    :param tuple(type) types_to_prune: Types to prune. Example: (otio.schema.Gap,...)
    """

    # deep copy everything
    mutable_object = copy.deepcopy(root)

    prune_list = set()

    header_list = [mutable_object]

    if isinstance(mutable_object, schema.Timeline):
        header_list.append(mutable_object.tracks)

    iter_list = header_list + list(mutable_object.find_children())

    for child in iter_list:
        if _safe_parent(child) is not None and _is_in(child.parent(), prune_list):
            prune_list.add(child)
            continue

        parent = None
        child_index = None
        if _safe_parent(child) is not None:
            child_index = child.parent().index(child)
            parent = child.parent()
            del child.parent()[child_index]

        # first try to prune
        if (types_to_prune and _isinstance_in(child, types_to_prune)):
            result = None
        # finally call the user function
        else:
            result = unary_filter_fn(child)

        if child is mutable_object:
            mutable_object = result

        if result is None:
            prune_list.add(child)
            continue

        if type(result) is not tuple:
            result = [result]

        if parent is not None:
            parent[child_index:child_index] = result

    return mutable_object


def _safe_parent(child):
    if hasattr(child, 'parent'):
        return child.parent()
    return None


def filtered_with_sequence_context(
    root,
    reduce_fn,
    types_to_prune=None,
):
    """Filter a deep copy of root (and children) with ``reduce_fn``.

    The ``reduce_fn`` must have this signature:

    .. currentmodule:: _

    .. py:function:: func(previous_item: typing.Any, current: typing.Any, next_item: typing.Any) -> list[typing.Any]  # noqa
        :noindex:

    1. Make a deep copy of root
    2. Starting with root, perform a depth first traversal
    3. For each item (including root):

       a. if types_to_prune is not None and item is an instance of a type
          in types_to_prune, prune it from the copy, continue.
       b. Otherwise, pass (prev, copy, and next) to reduce_fn. If ``reduce_fn``:

          I.   returns an object: add it to the copy, replacing original
          II.  returns a tuple: insert it into the list, replacing original
          III. returns None: prune it

          .. note:: ``reduce_fn`` is always passed objects from the original
                    deep copy, not what prior calls return. See below for examples

    4. If an item is pruned, do not traverse its children
    5. Return the new deep copy.

    Example 1 (filter)::

        >>> track = [A,B,C]
        >>> def fn(prev_item, thing, next_item):
        ...     if prev_item.name == A:
        ...         return D # some new clip
        ...     else:
        ...         return thing
        >>> filtered_with_sequence_context(track, fn) => [A,D,C]

        order of calls to fn:
            fn(None, A, B) => A
            fn(A, B, C) => D
            fn(B, C, D) => C # !! note that it was passed B instead of D.

    Example 2 (prune)::

        >>> track = [A,B,C]
        >>> def fn(prev_item, thing, next_item):
        ...    if prev_item.name == A:
        ...        return None # prune the clip
        ...   else:
        ...        return thing
        >>> filtered_with_sequence_context(track, fn) => [A,C]

        order of calls to fn:
            fn(None, A, B) => A
            fn(A, B, C) => None
            fn(B, C, D) => C # !! note that it was passed B instead of D.

    Example 3 (expand)::

        >>> def fn(prev_item, thing, next_item):
        ...     if prev_item.name == A:
        ...         return (D, E) # tuple of new clips
        ...     else:
        ...         return thing
        >>> filtered_with_sequence_context(track, fn) => [A, D, E, C]

         the order of calls to fn will be:
            fn(None, A, B) => A
            fn(A, B, C) => (D, E)
            fn(B, C, D) => C # !! note that it was passed B instead of D.

    :param SerializableObjectWithMetadata root: Object to filter on
    :param reduce_fn: Filter function
    :param tuple(type) types_to_prune: Types to prune. Example: (otio.schema.Gap,...)
    """

    # deep copy everything
    mutable_object = copy.deepcopy(root)

    prune_list = set()

    header_list = [mutable_object]

    if isinstance(mutable_object, schema.Timeline):
        header_list.append(mutable_object.tracks)

    iter_list = header_list + list(mutable_object.find_children())

    # expand to include prev, next when appropriate
    expanded_iter_list = []
    for child in iter_list:
        if _safe_parent(child) and isinstance(child.parent(), schema.Track):
            prev_item, next_item = child.parent().neighbors_of(child)
            expanded_iter_list.append((prev_item, child, next_item))
        else:
            expanded_iter_list.append((None, child, None))

    def _safe_name(x):
        return x.name if x else "<NONE>"

    for prev_item, child, next_item in expanded_iter_list:
        if _safe_parent(child) is not None and _is_in(child.parent(), prune_list):
            prune_list.add(child)
            continue

        parent = None
        child_index = None
        if _safe_parent(child) is not None:
            child_index = child.parent().index(child)
            parent = child.parent()
            del child.parent()[child_index]

        # first try to prune
        if types_to_prune and _isinstance_in(child, types_to_prune):
            result = None
        # finally call the user function
        else:
            result = reduce_fn(prev_item, child, next_item)

        if child is mutable_object:
            mutable_object = result

        if result is None:
            prune_list.add(child)
            continue

        if type(result) is not tuple:
            result = [result]

        if parent is not None:
            parent[child_index:child_index] = result

    return mutable_object



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

from .. core._core_utils import add_method
from .. import _otio


@add_method(_otio.Clip)
def __str__(self):
    return 'Clip("{}", {}, {}, {}, {}, {})'.format(
        self.name,
        self.media_reference,
        self.source_range,
        self.metadata,
        self.effects,
        self.markers
    )


@add_method(_otio.Clip)
def __repr__(self):
    return (
        'otio.schema.Clip('
        'name={}, '
        'media_reference={}, '
        'source_range={}, '
        'color={}, '
        'metadata={}, '
        'effects={}, '
        'markers={}'
        ')'.format(
            repr(self.name),
            repr(self.media_reference),
            repr(self.source_range),
            repr(self.color),
            repr(self.metadata),
            repr(self.effects),
            repr(self.markers)
        )
    )


@add_method(_otio.Clip)
def find_clips(self, search_range=None):
    yield self



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

import inspect
import collections

from .. import (
    core,
    exceptions,
    plugins,
    schemadef
)


@core.register_type
class SchemaDef(plugins.PythonPlugin):
    _serializable_label = "SchemaDef.1"

    def __init__(
        self,
        name=None,
        filepath=None,
    ):
        plugins.PythonPlugin.__init__(self, name, filepath)

    def module(self):
        """
        Return the module object for this schemadef plugin.
        If the module hasn't already been imported, it is imported and
        injected into the otio.schemadefs namespace as a side-effect.

        Redefines :py:meth:`.PythonPlugin.module`.
        """

        if not self._module:
            self._module = self._imported_module("schemadef")
            if self.name:
                schemadef._add_schemadef_module(self.name, self._module)

        return self._module

    def plugin_info_map(self):
        """Adds extra schemadef-specific information to call to the parent fn.
        """

        result = super().plugin_info_map()

        features = collections.OrderedDict()
        for name in dir(self.module()):
            thing = getattr(self.module(), name)
            if (
                not (
                    inspect.isclass(thing)
                    and issubclass(thing, core.SerializableObject)
                )
            ):
                continue
            features[name] = {
                "class": thing,
                "doc": inspect.getdoc(thing)
            }
            # @TODO: include fields

        result["SchemaDefs"] = features
        return result

    def __str__(self):
        return "SchemaDef({}, {})".format(
            repr(self.name),
            repr(self.filepath)
        )

    def __repr__(self):
        return (
            "otio.schema.SchemaDef("
            "name={}, "
            "filepath={}"
            ")".format(
                repr(self.name),
                repr(self.filepath)
            )
        )


def available_schemadef_names():
    """Return a string list of the available schemadefs."""

    return [str(sd.name) for sd in plugins.ActiveManifest().schemadefs]


def from_name(name):
    """Fetch the schemadef plugin object by the name of the schema directly."""

    try:
        return plugins.ActiveManifest().from_name(name, kind_list="schemadefs")
    except exceptions.NotSupportedError:
        raise exceptions.NotSupportedError(
            "schemadef not supported: {}, available: {}".format(
                name,
                available_schemadef_names()
            )
        )


def module_from_name(name):
    """Fetch the plugin's module by the name of the schemadef.

    Will load the plugin if it has not already been loaded.  Reading a file that
    contains the schemadef will also trigger a load of the plugin.
    """
    plugin = from_name(name)
    return plugin.module()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Console scripts for OpenTimelineIO

.. moduleauthor:: Contributors to the OpenTimelineIO project <otio-discussion@lists.aswf.io>
"""

# flake8: noqa

# in dependency hierarchy
from . import (
    otioconvert,
    otiocat,
    otiostat,
    otiotool,
    console_utils,
    autogen_serialized_datamodel,
    otiopluginfo,
)




# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

from .. core._core_utils import add_method
from .. import _otio


@add_method(_otio.ImageSequenceReference)
def __str__(self):
    return (
        'ImageSequenceReference('
        '"{}", "{}", "{}", {}, {}, {}, {}, {}, {}, {}, {})' .format(
            self.target_url_base,
            self.name_prefix,
            self.name_suffix,
            self.start_frame,
            self.frame_step,
            self.rate,
            self.frame_zero_padding,
            self.missing_frame_policy,
            self.available_range,
            self.available_image_bounds,
            self.metadata,
        )
    )


@add_method(_otio.ImageSequenceReference)
def __repr__(self):
    return (
        'ImageSequenceReference('
        'target_url_base={}, '
        'name_prefix={}, '
        'name_suffix={}, '
        'start_frame={}, '
        'frame_step={}, '
        'rate={}, '
        'frame_zero_padding={}, '
        'missing_frame_policy={}, '
        'available_range={}, '
        'available_image_bounds={}, '
        'metadata={}'
        ')' .format(
            repr(self.target_url_base),
            repr(self.name_prefix),
            repr(self.name_suffix),
            repr(self.start_frame),
            repr(self.frame_step),
            repr(self.rate),
            repr(self.frame_zero_padding),
            repr(self.missing_frame_policy),
            repr(self.available_range),
            repr(self.available_image_bounds),
            repr(self.metadata),
        )
    )


@add_method(_otio.ImageSequenceReference)
def frame_range_for_time_range(self, time_range):
    """Returns first and last frame numbers for
    the given time range in the reference.

    :rtype: tuple[int]
    :raises ValueError: if the provided time range is outside the available range.
    """
    return (
        self.frame_for_time(time_range.start_time),
        self.frame_for_time(time_range.end_time_inclusive())
    )


@add_method(_otio.ImageSequenceReference)
def abstract_target_url(self, symbol):
    """
    Generates a target url for a frame where ``symbol`` is used in place
    of the frame number. This is often used to generate wildcard target urls.
    """
    if not self.target_url_base.endswith("/"):
        base = self.target_url_base + "/"
    else:
        base = self.target_url_base

    return "{}{}{}{}".format(
        base, self.name_prefix, symbol, self.name_suffix
    )



#!/usr/bin/env python
#
# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Generate the CORE_VERSION_MAP for this version of OTIO"""

import argparse
import tempfile

import opentimelineio as otio


LABEL_MAP_TEMPLATE = """{{ "{label}",
      {{
{sv_map}
      }} }},
    // {{next}}"""
MAP_ITEM_TEMPLATE = '{indent}{{ "{key}", {value} }},'
INDENT = 10


def _parsed_args():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "-d",
        "--dryrun",
        default=False,
        action="store_true",
        help="write to stdout instead of printing to file."
    )
    parser.add_argument(
        "-l",
        "--label",
        default=otio.__version__,
        # @TODO - should we strip the .dev1 label?  that would probably be
        #         more consistent since we don't do sub-beta releases
        help="Version label to assign this schema map to."
    )
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        default=None,
        required=True,
        help="Path to CORE_VERSION_MAP.last.cpp"
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default=None,
        help="Path to where CORE_VERSION_MAP.cpp should be written to."
    )

    return parser.parse_args()


def generate_core_version_map(src_text, label, version_map):
    # turn the braces in the .cpp file into python-format template compatible
    # form ({{ }} where needed)
    src_text = src_text.replace("{", "{{").replace("}", "}}")
    src_text = src_text.replace("// {{next}}", "{next}")

    # iterate over the map and print the template out
    map_text = []
    for key, value in sorted(version_map.items()):
        map_text.append(
            MAP_ITEM_TEMPLATE.format(
                indent=' ' * INDENT,
                key=key,
                value=value
            )
        )
    map_text = '\n'.join(map_text)

    # assemble the result
    next_text = LABEL_MAP_TEMPLATE.format(label=label, sv_map=map_text)
    return src_text.format(label=label, next=next_text)


def main():
    args = _parsed_args()

    with open(args.input) as fi:
        input = fi.read()

    result = generate_core_version_map(
        input,
        args.label,
        otio.core.type_version_map()
    )

    if args.dryrun:
        print(result)
        return

    output = args.output
    if not output:
        output = tempfile.NamedTemporaryFile(
            'w',
            suffix="CORE_VERSION_MAP.cpp",
            delete=False
        ).name

    with open(output, 'w', newline="\n") as fo:
        fo.write(result)

    print(f"Wrote CORE_VERSION_MAP to: '{output}'.")


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""OTIOZ adapter - bundles otio files linked to local media

Takes as input an OTIO file that has media references which are all
ExternalReferences with target_urls to files with unique basenames that are
accessible through the file system and bundles those files and the otio file
into a single zip file with the suffix .otioz.  Can error out if files aren't
locally referenced or provide missing references

Can also extract the content.otio file from an otioz bundle for processing.

Note that OTIOZ files _always_ use the unix style path separator ('/'). This
ensures that regardless of which platform a bundle was created on, it can be
read on unix and windows platforms.
"""

import os
import zipfile

from .. import (
    exceptions,
    url_utils,
)

from . import (
    file_bundle_utils as utils,
    otio_json,
)

import pathlib


def read_from_file(
    filepath,
    # if provided, will extract contents of zip to this directory
    extract_to_directory=None,
):
    if not zipfile.is_zipfile(filepath):
        raise exceptions.OTIOError(f"Not a zipfile: {filepath}")

    if extract_to_directory:
        output_media_directory = os.path.join(
            extract_to_directory,
            utils.BUNDLE_DIR_NAME
        )

        if not os.path.exists(extract_to_directory):
            raise exceptions.OTIOError(
                f"Directory '{extract_to_directory}' does not exist, cannot"
                f" unpack otioz there."
            )

        if os.path.exists(output_media_directory):
            raise exceptions.OTIOError(
                f"Error: '{output_media_directory}' already exists on disk, "
                f"cannot overwrite while unpacking OTIOZ file '{filepath}'."
            )

    with zipfile.ZipFile(filepath, 'r') as zi:
        result = otio_json.read_from_string(
            zi.read(utils.BUNDLE_PLAYLIST_PATH)
        )

        if extract_to_directory:
            zi.extractall(extract_to_directory)

    return result


def write_to_file(
    input_otio,
    filepath,
    # see documentation in file_bundle_utils for more information on the
    # media_policy
    media_policy=utils.MediaReferencePolicy.ErrorIfNotFile,
    dryrun=False
):
    if os.path.exists(filepath):
        raise exceptions.OTIOError(
            f"'{filepath}' exists, will not overwrite."
        )

    # general algorithm for the file bundle adapters:
    # -------------------------------------------------------------------------
    # - build file manifest (list of paths to files on disk that will be put
    #   into the archive)
    # - build a mapping of path to file on disk to url to put into the media
    #   reference in the result
    # - relink the media references to point at the final location inside the
    #   archive
    # - build the resulting structure (zip file, directory)
    # -------------------------------------------------------------------------

    result_otio, path_to_mr_map = utils._prepped_otio_for_bundle_and_manifest(
        input_otio,
        media_policy,
        "OTIOZ"
    )

    # dryrun reports the total size of files
    if dryrun:
        return utils._total_file_size_of(path_to_mr_map.keys())

    abspath_to_output_path_map = {}

    # relink all the media references to their target paths
    for abspath, references in path_to_mr_map.items():
        target = os.path.join(utils.BUNDLE_DIR_NAME, os.path.basename(abspath))

        # conform to posix style paths inside the bundle, so that they are
        # portable between windows and *nix style environments
        final_path = str(pathlib.Path(target).as_posix())

        # cache the output path
        abspath_to_output_path_map[abspath] = final_path

        for mr in references:
            # author the final_path in url form into the target_url
            mr.target_url = url_utils.url_from_filepath(final_path)

    # write the otioz file to the temp directory
    otio_str = otio_json.write_to_string(result_otio)

    with zipfile.ZipFile(filepath, mode='w') as target:
        # write the version file (compressed)
        target.writestr(
            utils.BUNDLE_VERSION_FILE,
            utils.BUNDLE_VERSION,
            # XXX: OTIOZ was introduced when python 2.7 was still a supported
            #      platform. The newer algorithms, like BZIP2 and LZMA, are not
            #      available in python2, so it uses the zlib based
            #      ZIP_DEFLATED.  Now that OTIO is Python3+, this could switch
            #      to using BZIP2 or LZMA instead... with the caveat that this
            #      would make OTIOZ files incompatible with python 2 based OTIO
            #      installs.
            #
            #      For example, if we used ZIP_LZMA, then otio release v0.15
            #      would still be able to open these files as long as the
            #      python interpreter was version 3+.
            compress_type=zipfile.ZIP_DEFLATED
        )

        # write the OTIO (compressed)
        target.writestr(
            utils.BUNDLE_PLAYLIST_PATH,
            otio_str,
            # XXX: See comment above about ZIP_DEFLATED vs other algorithms
            compress_type=zipfile.ZIP_DEFLATED
        )

        # write the media (uncompressed)
        for src, dst in abspath_to_output_path_map.items():
            target.write(src, dst, compress_type=zipfile.ZIP_STORED)

    return



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

from .. core._core_utils import add_method
from .. import _otio


@add_method(_otio.Box2d)
def __str__(self):
    return 'Box2d({}, {})'.format(
        self.min,
        self.max
    )


@add_method(_otio.Box2d)
def __repr__(self):
    return (
        'otio.schema.Box2d('
        'min={}, '
        'max={}'
        ')'.format(
            repr(self.min),
            repr(self.max),
        )
    )



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

r"""
MediaLinker plugins fire after an adapter has read a file in order to
produce :class:`.MediaReference`\s that point at valid, site specific media.

They expose a ``link_media_reference`` function with the signature:

.. py:function:: link_media_reference(in_clip: opentimelineio.schema.Clip) -> opentimelineio.core.MediaReference  # noqa
   :noindex:

   Example link_media_reference function.

To get context information, they can inspect the metadata on the clip and on
the media reference. The :meth:`.Composable.parent` method can be used to find the containing
track if metadata is stored there.
"""

import os
import inspect

from . import (
    exceptions,
    plugins,
    core,
)


class MediaLinkingPolicy:
    """Enum describing different media linker policies"""
    DoNotLinkMedia = "__do_not_link_media"
    ForceDefaultLinker = "__default"


# @TODO: wrap this up in the plugin system somehow?  automatically generate?
def available_media_linker_names():
    """Return a string list of the available media linker plugins."""

    return [str(adp.name) for adp in plugins.ActiveManifest().media_linkers]


def from_name(name):
    """Fetch the media linker object by the name of the adapter directly."""

    if name == MediaLinkingPolicy.ForceDefaultLinker or not name:
        name = os.environ.get("OTIO_DEFAULT_MEDIA_LINKER", None)

    if not name:
        return None

    # @TODO: make this handle the enums
    try:
        return plugins.ActiveManifest().from_name(
            name,
            kind_list="media_linkers"
        )
    except exceptions.NotSupportedError:
        raise exceptions.NotSupportedError(
            "media linker not supported: {}, available: {}".format(
                name,
                available_media_linker_names()
            )
        )


def default_media_linker():
    try:
        return os.environ['OTIO_DEFAULT_MEDIA_LINKER']
    except KeyError:
        raise exceptions.NoDefaultMediaLinkerError(
            "No default Media Linker set in $OTIO_DEFAULT_MEDIA_LINKER"
        )


def linked_media_reference(
    target_clip,
    media_linker_name=MediaLinkingPolicy.ForceDefaultLinker,
    media_linker_argument_map=None
):
    media_linker = from_name(media_linker_name)

    if not media_linker:
        return target_clip

    # @TODO: connect this argument map up to the function call through to the
    #        real linker
    if not media_linker_argument_map:
        media_linker_argument_map = {}

    return media_linker.link_media_reference(
        target_clip,
        media_linker_argument_map
    )


@core.register_type
class MediaLinker(plugins.PythonPlugin):
    _serializable_label = "MediaLinker.1"

    def __init__(
        self,
        name=None,
        filepath=None,
    ):
        super().__init__(name, filepath)

    def link_media_reference(self, in_clip, media_linker_argument_map=None):
        media_linker_argument_map = media_linker_argument_map or {}

        return self._execute_function(
            "link_media_reference",
            in_clip=in_clip,
            media_linker_argument_map=media_linker_argument_map
        )

    def is_default_linker(self):
        return os.environ.get("OTIO_DEFAULT_MEDIA_LINKER", "") == self.name

    def plugin_info_map(self):
        """Adds extra adapter-specific information to call to the parent fn."""

        result = super().plugin_info_map()

        fn_doc = inspect.getdoc(self.module().link_media_reference)
        if fn_doc:
            mod_doc = [result['doc'], ""]
            mod_doc.append(fn_doc)
            result["doc"] = "\n".join(mod_doc)

        if self.is_default_linker():
            result["** CURRENT DEFAULT MEDIA LINKER"] = True

        return result

    def __str__(self):
        return "MediaLinker({}, {})".format(
            repr(self.name),
            repr(self.filepath)
        )

    def __repr__(self):
        return (
            "otio.media_linker.MediaLinker("
            "name={}, "
            "filepath={}"
            ")".format(
                repr(self.name),
                repr(self.filepath)
            )
        )



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Plugin system for OTIO"""

# flake8: noqa

from .python_plugin import (
    plugin_info_map,
    PythonPlugin,
)

from .manifest import (
    manifest_from_file,
    ActiveManifest,
)



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

from .. core._core_utils import add_method
from .. import _otio


@add_method(_otio.Marker)
def __str__(self):
    return "Marker({}, {}, {})".format(
        str(self.name),
        str(self.marked_range),
        str(self.metadata),
    )


@add_method(_otio.Marker)
def __repr__(self):
    return (
        "otio.schema.Marker("
        "name={}, "
        "marked_range={}, "
        "metadata={}"
        ")".format(
            repr(self.name),
            repr(self.marked_range),
            repr(self.metadata),
        )
    )



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

def _add_schemadef_module(name, mod):
    """Insert a new module name and module object into schemadef namespace."""
    ns = globals()  # the namespace dict of the schemadef package
    ns[name] = mod



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Exception classes for OpenTimelineIO"""
from . _otio import ( # noqa
    OTIOError,
    NotAChildError,
    UnsupportedSchemaError,
    CannotComputeAvailableRangeError
)

__all__ = [
    'OTIOError',
    'NotAChildError',
    'CannotComputeAvailableRangeError',
    'UnsupportedSchemaError',
    'CouldNotReadFileError',
    'NoKnownAdapterForExtensionError',
    'ReadingNotSupportedError',
    'WritingNotSupportedError',
    'NotSupportedError',
    'InvalidSerializableLabelError',
    'AdapterDoesntSupportFunctionError',
    'InstancingNotAllowedError',
    'TransitionFollowingATransitionError',
    'MisconfiguredPluginError',
    'CannotTrimTransitionsError',
    'NoDefaultMediaLinkerError'
]


class CouldNotReadFileError(OTIOError):
    pass


class NoKnownAdapterForExtensionError(OTIOError):
    pass


class ReadingNotSupportedError(OTIOError):
    pass


class WritingNotSupportedError(OTIOError):
    pass


class NotSupportedError(OTIOError):
    pass


class InvalidSerializableLabelError(OTIOError):
    pass


class AdapterDoesntSupportFunctionError(OTIOError):
    pass


class InstancingNotAllowedError(OTIOError):
    pass


class TransitionFollowingATransitionError(OTIOError):
    pass


class MisconfiguredPluginError(OTIOError):
    pass


class CannotTrimTransitionsError(OTIOError):
    pass


class NoDefaultMediaLinkerError(OTIOError):
    pass


class InvalidEnvironmentVariableError(OTIOError):
    pass



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

import types
import collections.abc
import copy

from .. import (
    _otio,
)
from .. _otio import (
    SerializableObject,
    AnyDictionary,
    AnyVector,
    PyAny
)


SUPPORTED_VALUE_TYPES = (
    "int",
    "float",
    "str",
    "bool",
    "list",
    "dictionary",
    "opentime.RationalTime",
    "opentime.TimeRange",
    "opentime.TimeTransform",
    "opentimelineio.core.Color",
    "opentimelineio.core.SerializableObject"
)


def _is_str(v):
    return isinstance(v, str)


def _is_nonstring_sequence(v):
    return isinstance(v, collections.abc.Sequence) and not _is_str(v)


def _value_to_any(value, ids=None):
    if isinstance(value, PyAny):
        return value

    if isinstance(value, SerializableObject):
        return PyAny(value)
    if isinstance(value, collections.abc.Mapping):
        if ids is None:
            ids = set()
        d = AnyDictionary()
        for (k, v) in value.items():
            if not _is_str(k):
                raise ValueError(f"key '{k}' is not a string")
            if id(v) in ids:
                raise ValueError(
                    "circular reference converting dictionary to C++ datatype"
                )
            ids.add(id(v))
            d[k] = _value_to_any(v, ids)
            ids.discard(id(v))
        return PyAny(d)
    elif _is_nonstring_sequence(value):
        if ids is None:
            ids = set()
        vec = AnyVector()
        for v in value:
            if id(v) in ids:
                raise ValueError(
                    "circular reference converting dictionary to C++ datatype"
                )
            ids.add(id(v))
            vec.append(_value_to_any(v, ids))
            ids.discard(id(v))
        return PyAny(vec)
    else:
        try:
            return PyAny(value)
        except TypeError:
            # raise an OTIO-specific error
            raise TypeError(
                "A value of type '{}' is incompatible with OpenTimelineIO. "
                "OpenTimelineIO only supports the following value types in "
                "AnyDictionary containers (like the .metadata dictionary): "
                "{}.".format(
                    type(value),
                    SUPPORTED_VALUE_TYPES,
                )
            )
        except RuntimeError:
            # communicate about integer range first
            biginttype = int
            if isinstance(value, biginttype):
                raise ValueError(
                    "A value of {} is outside of the range of integers that "
                    "OpenTimelineIO supports, [{}, {}], which is the range of "
                    "C++ int64_t.".format(
                        value,
                        -9223372036854775808,
                        9223372036854775807,
                    )
                )

            # general catch all for invalid type
            raise ValueError(
                "The value '{}' of type '{}' is incompatible with OpenTimelineIO. "
                "OpenTimelineIO only supports the following value types in "
                "AnyDictionary containers (like the .metadata dictionary): "
                "{}.".format(
                    value,
                    type(value),
                    SUPPORTED_VALUE_TYPES,
                )
            )


_marker_ = object()


def _add_mutable_mapping_methods(mapClass):
    def __setitem__(self, key, item):
        self.__internal_setitem__(key, _value_to_any(item))

    def __str__(self):
        return str(dict(self))

    def __repr__(self):
        return repr(dict(self))

    def setdefault(self, key, default_value):
        if key in self:
            return self[key]
        else:
            self[key] = default_value
            return self[key]

    def pop(self, key, default=_marker_):
        try:
            value = self[key]
        except KeyError:
            if default is _marker_:
                raise
            return default
        else:
            del self[key]
            return value

    def __copy__(self):
        m = mapClass()
        m.update({k: v for (k, v) in self.items()})
        return m

    def __deepcopy__(self, memo):
        m = mapClass()
        m.update({k: copy.deepcopy(v, memo) for (k, v) in self.items()})
        return m

    collections.abc.MutableMapping.register(mapClass)
    mapClass.__setitem__ = __setitem__
    mapClass.__str__ = __str__
    mapClass.__repr__ = __repr__

    seen = set()
    for klass in (collections.abc.MutableMapping, collections.abc.Mapping):
        for name in klass.__dict__.keys():
            if name in seen:
                continue

            seen.add(name)
            func = getattr(klass, name)
            if (
                    isinstance(func, types.FunctionType)
                    and name not in klass.__abstractmethods__
            ):
                setattr(mapClass, name, func)
                if name.startswith('__') or name.endswith('__'):  # noqa
                    continue

                # Hide the method frm Sphinx doc.
                # See https://www.sphinx-doc.org/en/master/usage/restructuredtext/domains.html#info-field-lists  # noqa
                # __doc__ will be None on a Windows embeddable package where the built-in modules are provided  # noqa
                # as pyc files which do not include the docstrings.
                if getattr(mapClass, name).__doc__ is not None:
                    getattr(mapClass, name).__doc__ += "\n\n:meta private:"

    mapClass.setdefault = setdefault
    mapClass.pop = pop
    mapClass.__copy__ = __copy__
    mapClass.__deepcopy__ = __deepcopy__


def _add_mutable_sequence_methods(
        sequenceClass,
        conversion_func=None,
        side_effecting_insertions=False
):
    def noop(x):
        return x

    if not conversion_func:
        conversion_func = noop

    def __add__(self, other):
        if isinstance(other, list):
            return list(self) + other
        elif _is_nonstring_sequence(other):
            return list(self) + list(other)
        else:
            raise TypeError(
                f"Cannot add types '{type(self)}' and '{type(other)}'"
            )

    def __radd__(self, other):
        return self.__add__(other)

    def __str__(self):
        return str(list(self))

    def __repr__(self):
        return repr(list(self))

    def __getitem__(self, index):
        if isinstance(index, slice):
            indices = index.indices(len(self))
            return [self.__internal_getitem__(i) for i in range(*indices)]
        else:
            return self.__internal_getitem__(index)

    # This has to handle slicing
    def __setitem__(self, index, item):
        if not isinstance(index, slice):
            self.__internal_setitem__(index, conversion_func(item))
        else:
            if not isinstance(item, collections.abc.Iterable):
                raise TypeError("can only assign an iterable")

            indices = range(*index.indices(len(self)))

            if index.step in (1, None):
                if (
                        not side_effecting_insertions
                        and isinstance(item, collections.abc.MutableSequence)
                        and len(item) == len(indices)
                ):
                    for i0, i in enumerate(indices):
                        self.__internal_setitem__(i, conversion_func(item[i0]))
                else:
                    if side_effecting_insertions:
                        cached_items = list(self)

                    for i in reversed(indices):
                        self.__internal_delitem__(i)
                    insertion_index = 0 if index.start is None else index.start

                    if not side_effecting_insertions:
                        for e in item:
                            self.__internal_insert(insertion_index, e)
                            insertion_index += 1
                    else:
                        try:
                            for e in item:
                                self.__internal_insert(insertion_index, e)
                                insertion_index += 1
                        except Exception as e:
                            # restore the old state
                            while len(self):
                                self.pop()
                            self.extend(cached_items)
                            raise e
            else:
                if not isinstance(item, collections.abc.Sequence):
                    raise TypeError("can only assign a sequence")
                if len(item) != len(indices):
                    raise ValueError(
                        "attempt to assign sequence of size {} to extended "
                        "slice of size {}".format(len(item), len(indices))
                    )
                if not side_effecting_insertions:
                    for i, e in enumerate(item):
                        self.__internal_setitem__(indices[i], conversion_func(e))
                else:
                    cached_items = list(self)
                    for index in reversed(indices):
                        self.__internal_del_item__(index)
                    try:
                        for i, e in enumerate(item):
                            self.__internal_insert(indices[i], e)
                    except Exception as e:
                        # restore the old state
                        while len(self):
                            self.pop()
                        self.extend(cached_items)
                        raise e

    # This has to handle slicing
    def __delitem__(self, index):
        if not isinstance(index, slice):
            self.__internal_delitem__(index)
        else:
            for i in reversed(range(*index.indices(len(self)))):
                self.__delitem__(i)

    def insert(self, index, item):
        self.__internal_insert(
            index, conversion_func(item)
            if conversion_func else item
        )

    collections.abc.MutableSequence.register(sequenceClass)
    sequenceClass.__radd__ = __radd__
    sequenceClass.__add__ = __add__
    sequenceClass.__getitem__ = __getitem__
    sequenceClass.__setitem__ = __setitem__
    sequenceClass.__delitem__ = __delitem__
    sequenceClass.insert = insert
    sequenceClass.__str__ = __str__
    sequenceClass.__repr__ = __repr__

    seen = set()
    for klass in (collections.abc.MutableSequence, collections.abc.Sequence):
        for name in klass.__dict__.keys():
            if name not in seen:
                seen.add(name)
                func = getattr(klass, name)
                if (
                        isinstance(func, types.FunctionType)
                        and name not in klass.__abstractmethods__
                        and not hasattr(sequenceClass, name)
                ):
                    setattr(sequenceClass, name, func)
                    if name.startswith('__') or name.endswith('__'):
                        continue

                    # Hide the method frm Sphinx doc.
                    # See https://www.sphinx-doc.org/en/master/usage/restructuredtext/domains.html#info-field-lists  # noqa
                    # __doc__ will be None on a Windows embeddable package where the built-in modules are provided  # noqa
                    # as pyc files which do not include the docstrings.
                    if getattr(sequenceClass, name).__doc__ is not None:
                        getattr(sequenceClass, name).__doc__ += "\n\n:meta private:"

    if not issubclass(sequenceClass, SerializableObject):
        def __copy__(self):
            v = sequenceClass()
            v.extend(e for e in self)
            return v

        def __deepcopy__(self, memo=None):
            v = sequenceClass()
            v.extend(copy.deepcopy(e, memo) for e in self)
            return v

        sequenceClass.__copy__ = __copy__
        sequenceClass.__deepcopy__ = __deepcopy__


_add_mutable_mapping_methods(AnyDictionary)
_add_mutable_sequence_methods(AnyVector, conversion_func=_value_to_any)
_add_mutable_sequence_methods(_otio.MarkerVector)
_add_mutable_sequence_methods(_otio.EffectVector)
_add_mutable_sequence_methods(_otio.Composition, side_effecting_insertions=True)
_add_mutable_sequence_methods(_otio.SerializableCollection)


def __setattr__(self, key, value):
    super(SerializableObject, self).__setattr__(key, value)
    _otio.install_external_keepalive_monitor(self, True)


SerializableObject.__setattr__ = __setattr__


# Decorator that adds a function into a class.
def add_method(cls):
    def decorator(func):
        setattr(cls, func.__name__, func)
    return decorator


@add_method(SerializableObject)
def deepcopy(self, *args, **kwargs):
    return self.clone()


@add_method(SerializableObject)
def __deepcopy__(self, *args, **kwargs):
    return self.clone()


@add_method(SerializableObject)
def __copy__(self, *args, **kwargs):
    raise ValueError("SerializableObjects may not be shallow copied.")



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""OTIOD adapter - bundles otio files linked to local media in a directory

Takes as input an OTIO file that has media references which are all
ExternalReferences with target_urls to files with unique basenames that are
accessible through the file system and bundles those files and the otio file
into a single directory named with a suffix of .otiod.
"""

import os
import shutil

from . import (
    file_bundle_utils as utils,
    otio_json,
)

from .. import (
    exceptions,
    url_utils,
)

import pathlib
import urllib.parse as urlparse


def read_from_file(
    filepath,
    # convert the media_reference paths to absolute paths
    absolute_media_reference_paths=False,
):
    result = otio_json.read_from_file(
        os.path.join(filepath, utils.BUNDLE_PLAYLIST_PATH)
    )

    if not absolute_media_reference_paths:
        return result

    for cl in result.find_clips():
        try:
            source_fpath = cl.media_reference.target_url
        except AttributeError:
            continue

        rel_path = urlparse.urlparse(source_fpath).path
        new_fpath = url_utils.url_from_filepath(
            os.path.join(filepath, rel_path)
        )

        cl.media_reference.target_url = new_fpath

    return result


def write_to_file(
    input_otio,
    filepath,
    # see documentation in file_bundle_utils for more information on the
    # media_policy
    media_policy=utils.MediaReferencePolicy.ErrorIfNotFile,
    dryrun=False
):

    if os.path.exists(filepath):
        raise exceptions.OTIOError(
            f"'{filepath}' exists, will not overwrite."
        )

    if not os.path.exists(os.path.dirname(filepath)):
        raise exceptions.OTIOError(
            f"Directory '{os.path.dirname(filepath)}' does not exist, cannot"
            f" create '{filepath}'."
        )

    if not os.path.isdir(os.path.dirname(filepath)):
        raise exceptions.OTIOError(
            f"'{os.path.dirname(filepath)}' is not a directory, cannot create"
            f" '{filepath}'."
        )

    # general algorithm for the file bundle adapters:
    # -------------------------------------------------------------------------
    # - build file manifest (list of paths to files on disk that will be put
    #   into the archive)
    # - build a mapping of path to file on disk to url to put into the media
    #   reference in the result
    # - relink the media references to point at the final location inside the
    #   archive
    # - build the resulting structure (zip file, directory)
    # -------------------------------------------------------------------------

    result_otio, path_to_mr_map = utils._prepped_otio_for_bundle_and_manifest(
        input_otio,
        media_policy,
        "OTIOD"
    )

    # dryrun reports the total size of files
    if dryrun:
        return utils._total_file_size_of(path_to_mr_map.keys())

    abspath_to_output_path_map = {}

    # relink all the media references to their target paths
    for abspath, references in path_to_mr_map.items():
        target = os.path.join(
            filepath,
            utils.BUNDLE_DIR_NAME,
            os.path.basename(abspath)
        )

        # conform to posix style paths inside the bundle, so that they are
        # portable between windows and *nix style environments
        final_path = str(pathlib.Path(target).as_posix())

        # cache the output path
        abspath_to_output_path_map[abspath] = final_path

        for mr in references:
            # author the relative path from the root of the bundle in url
            # form into the target_url
            mr.target_url = url_utils.url_from_filepath(
                os.path.relpath(final_path, filepath)
            )

    os.mkdir(filepath)

    otio_json.write_to_file(
        result_otio,
        os.path.join(filepath, utils.BUNDLE_PLAYLIST_PATH)
    )

    # write the media files
    os.mkdir(os.path.join(filepath, utils.BUNDLE_DIR_NAME))
    for src, dst in abspath_to_output_path_map.items():
        shutil.copyfile(src, dst)

    return



#!/usr/bin/env python
#
# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Utility assertions for OTIO Unit tests."""

import re

from . import (
    adapters
)


class OTIOAssertions:
    def assertJsonEqual(self, known, test_result):
        """Convert to json and compare that (more readable)."""
        self.maxDiff = None

        known_str = adapters.write_to_string(known, 'otio_json')
        test_str = adapters.write_to_string(test_result, 'otio_json')

        def strip_trailing_decimal_zero(s):
            return re.sub(r'"(value|rate)": (\d+)\.0', r'"\1": \2', s)

        self.assertMultiLineEqual(
            strip_trailing_decimal_zero(known_str),
            strip_trailing_decimal_zero(test_str)
        )

    def assertIsOTIOEquivalentTo(self, known, test_result):
        """Test using the 'is equivalent to' method on SerializableObject"""

        self.assertTrue(known.is_equivalent_to(test_result))



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Common utilities used by the file bundle adapters (otiod and otioz)."""

import os
import copy

from .. import (
    exceptions,
    schema,
    url_utils,
)

import urllib


# versioning
BUNDLE_VERSION = "1.0.0"
BUNDLE_VERSION_FILE = "version.txt"

# other variables
BUNDLE_PLAYLIST_PATH = "content.otio"
BUNDLE_DIR_NAME = "media"


class NotAFileOnDisk(exceptions.OTIOError):
    pass


class MediaReferencePolicy:
    ErrorIfNotFile = "ErrorIfNotFile"
    MissingIfNotFile = "MissingIfNotFile"
    AllMissing = "AllMissing"


def reference_cloned_and_missing(orig_mr, reason_missing):
    """Replace orig_mr with a missing reference with the same metadata.

    Also adds original_target_url and missing_reference_because fields.
    """

    orig_mr = copy.deepcopy(orig_mr)
    media_reference = schema.MissingReference()
    media_reference.__dict__ = orig_mr.__dict__
    media_reference.metadata['missing_reference_because'] = reason_missing
    media_reference.metadata['original_target_url'] = orig_mr.target_url

    return media_reference


def _guarantee_unique_basenames(path_list, adapter_name):
    # walking across all unique file references, guarantee that all the
    # basenames are unique
    basename_to_source_fn = {}
    for fn in path_list:
        new_basename = os.path.basename(fn)
        if new_basename in basename_to_source_fn:
            raise exceptions.OTIOError(
                f"Error: the {adapter_name} adapter requires that the media"
                f" files have unique basenames.  File '{fn}' and"
                f" '{basename_to_source_fn[new_basename]}' have matching"
                f" basenames of: '{new_basename}'"
            )
        basename_to_source_fn[new_basename] = fn


def _prepped_otio_for_bundle_and_manifest(
    input_otio,    # otio to process
    media_policy,  # how to handle media references (see: MediaReferencePolicy)
    adapter_name,  # just for error messages
):
    """ Create a new OTIO based on input_otio that has had media references
    replaced according to the media_policy.  Return that new OTIO and a
    mapping of all the absolute file paths (not URLs) to be used in the bundle,
    mapped to MediaReferences associated with those files.  Media references in
    the OTIO will be relinked by the adapters to point to their output
    locations.

    The otio[dz] adapters use this function to do further relinking and build
    their bundles.

    This is considered an internal API.

    media_policy is expected to be of type MediaReferencePolicy.
    """

    # make sure the incoming OTIO isn't edited
    result_otio = copy.deepcopy(input_otio)

    path_to_reference_map = {}
    invalid_files = set()

    # result_otio is manipulated in place
    for cl in result_otio.find_clips():
        if media_policy == MediaReferencePolicy.AllMissing:
            cl.media_reference = reference_cloned_and_missing(
                cl.media_reference,
                f"{media_policy} specified as the MediaReferencePolicy"
            )
            continue

        try:
            target_url = cl.media_reference.target_url
        except AttributeError:
            # not an ExternalReference, ignoring it.
            continue

        parsed_url = urllib.parse.urlparse(target_url)

        # ensure that the urlscheme is either "file" or ""
        # file means "absolute path"
        # "" is interpreted as a relative path, relative to cwd of the python
        # process
        if parsed_url.scheme not in ("file", ""):
            if media_policy is MediaReferencePolicy.ErrorIfNotFile:
                raise NotAFileOnDisk(
                    f"The {adapter_name} adapter only works with media"
                    " reference target_url attributes that begin with 'file:'."
                    f"  Got a target_url of:  '{target_url}'"
                )
            if media_policy is MediaReferencePolicy.MissingIfNotFile:
                cl.media_reference = reference_cloned_and_missing(
                    cl.media_reference,
                    "target_url is not a file scheme url (start with url:)"
                )
                continue

        # get an absolute path to the target file
        target_file = os.path.abspath(url_utils.filepath_from_url(target_url))

        # if the file hasn't already been checked
        if (
            target_file not in path_to_reference_map
            and target_file not in invalid_files
            and (
                not os.path.exists(target_file)
                or not os.path.isfile(target_file)
            )
        ):
            invalid_files.add(target_file)

        if target_file in invalid_files:
            if media_policy is MediaReferencePolicy.ErrorIfNotFile:
                raise NotAFileOnDisk(target_file)
            if media_policy is MediaReferencePolicy.MissingIfNotFile:
                cl.media_reference = reference_cloned_and_missing(
                    cl.media_reference,
                    "target_url target is not a file or does not exist"
                )

                # do not need to relink it in the future or add this target to
                # the manifest, because the path is either not a file or does
                # not exist.
                continue

        # add the media reference to the list of references that point at this
        # file, they will need to be relinked
        path_to_reference_map.setdefault(target_file, []).append(
            cl.media_reference
        )

    _guarantee_unique_basenames(path_to_reference_map.keys(), adapter_name)

    return result_otio, path_to_reference_map


def _total_file_size_of(filepaths):
    fsize = 0
    for fn in filepaths:
        fsize += os.path.getsize(fn)
    return fsize



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

from . _core_utils import add_method
from .. import _otio


@add_method(_otio.Composition)
def __str__(self):
    return "{}({}, {}, {}, {})".format(
        self.__class__.__name__,
        str(self.name),
        str(list(self)),
        str(self.source_range),
        str(self.metadata)
    )


@add_method(_otio.Composition)
def __repr__(self):
    return (
        "otio.{}.{}("
        "name={}, "
        "children={}, "
        "source_range={}, "
        "color={}, "
        "metadata={}"
        ")".format(
            "core" if self.__class__ is _otio.Composition else "schema",
            self.__class__.__name__,
            repr(self.name),
            repr(list(self)),
            repr(self.source_range),
            repr(self.color),
            repr(self.metadata)
        )
    )



#!/usr/bin/env python
#
# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project


"""Generates documentation of the serialized data model for OpenTimelineIO."""

import argparse
import inspect
import json
import tempfile
import sys
import textwrap

import io

import opentimelineio as otio


DOCUMENT_HEADER = """# Serialized Data Documentation

This documents all the OpenTimelineIO classes that serialize to and from JSON,
omitting SchemaDef plugins. This document is automatically generated by
running:

`src/py-opentimelineio/opentimelineio/console/autogen_serialized_datamodel.py`

or by running:

`make doc-model`

It is part of the unit tests suite and should be updated whenever the schema
changes.  If it needs to be updated and this file regenerated, run:

`make doc-model-update`

# Class Documentation

"""

FIELDS_ONLY_HEADER = """# Serialized Data (Fields Only)

This document is a list of all the OpenTimelineIO classes that serialize to and
from JSON, omitting plugins classes and docstrings.

This document is automatically generated by running:

`src/py-opentimelineio/opentimelineio/console/autogen_serialized_datamodel.py`

or by running:

`make doc-model`

It is part of the unit tests suite and should be updated whenever the schema
changes.  If it needs to be updated and this file regenerated, run:

`make doc-model-update`


# Classes

"""

CLASS_HEADER_WITH_DOCS = """
### {classname}

*full module path*: `{modpath}`

*documentation*:

```
{docstring}
```

parameters:
"""

CLASS_HEADER_ONLY_FIELDS = """
### {classname}

parameters:
"""

MODULE_HEADER = """
## Module: {modname}
"""

PROP_HEADER = """- *{propkey}*: {prophelp}
"""

# @TODO: having type information here would be awesome
PROP_HEADER_NO_HELP = """- *{propkey}*
"""

# three ways to try and get the property + docstring
PROP_FETCHERS = (
    lambda cl, k: inspect.getdoc(getattr(cl, k)),
    lambda cl, k: inspect.getdoc(getattr(cl, "_" + k)),
    lambda cl, k: inspect.getdoc(getattr(cl(), k)) and "" or "",
)


def _parsed_args():
    """ parse commandline arguments with argparse """

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-d",
        "--dryrun",
        action="store_true",
        default=False,
        help="Dryrun mode - print out instead of perform actions"
    )
    group.add_argument(
        "-o",
        "--output",
        type=str,
        default=None,
        help="Update the baseline with the current version"
    )

    return parser.parse_args()


# things to skip.  These are typically internal/private classes that don't need
# to be considered "Public Schema".
SKIP_CLASSES = [
    otio.core.SerializableObject,
    otio._otio.UnknownSchema,
    otio._otio.TestObject,
]
SKIP_KEYS = [
    "OTIO_SCHEMA",  # not data, just for the backing format
    "children",     # children are stored by container objects, implicitly
]
SKIP_MODULES = [
    "opentimelineio.schemadef",  # because these are plugins
    "opentimelineio._otio",      # C++ .so, but customers should use the full
    "opentimelineio._opentime",  # python wrapped modules (otio.schema etc)
]


def _generate_model_for_module(mod, classes, modules):
    modules.add(mod)

    # fetch the classes from this module
    serializable_classes = [
        thing for thing in mod.__dict__.values()
        if (
            inspect.isclass(thing)
            and thing not in classes
            and issubclass(thing, otio.core.SerializableObject)
            or thing in (
                otio.opentime.RationalTime,
                otio.opentime.TimeRange,
                otio.opentime.TimeTransform,
                otio.core.Color
            )
        )
    ]

    # serialize/deserialize the classes to capture their serialized parameters
    model = {}
    for cl in serializable_classes:
        if cl in SKIP_CLASSES:
            continue

        model[cl] = {}
        field_dict = json.loads(otio.adapters.otio_json.write_to_string(cl()))
        for k in field_dict.keys():
            if k in SKIP_KEYS:
                continue

            for fetcher in PROP_FETCHERS:
                try:
                    # Serialized fields are almost always properties, but skip
                    # over those that are not
                    model[cl][k] = fetcher(cl, k) if isinstance(
                        getattr(cl, k), property) else ""
                    break
                except AttributeError:
                    pass
            else:
                sys.stderr.write(f"ERROR: could not fetch property: {k}")

        # Stashing the OTIO_SCHEMA back into the dictionary since the
        # documentation uses this information in its header.
        model[cl]["OTIO_SCHEMA"] = field_dict["OTIO_SCHEMA"]

    classes.update(model)

    # find new modules to recurse into
    new_mods = sorted(
        (
            thing for thing in mod.__dict__.values()
            if (
                inspect.ismodule(thing)
                and thing not in modules
                and "opentimelineio" in thing.__name__
                and all(not thing.__name__.startswith(t) for t in SKIP_MODULES)
            )
        ),
        key=lambda mod: str(mod)
    )

    # recurse into the new modules and update the classes and modules values
    [_generate_model_for_module(m, classes, modules) for m in new_mods]


def _generate_model():
    classes = {}
    modules = set()
    _generate_model_for_module(otio, classes, modules)
    return classes


def _search_mod_recursively(cl, mod_to_search, already_searched):
    if cl in mod_to_search.__dict__.values():
        return mod_to_search.__name__

    child_modules = (
        m
        for m in mod_to_search.__dict__.values()
        if inspect.ismodule(m)
    )

    for submod in child_modules:
        if submod in already_searched:
            continue
        already_searched.add(submod)
        result = _search_mod_recursively(cl, submod, already_searched)
        if result is not None:
            return result

    return None


def _remap_to_python_modules(cl):
    """Find the module containing the python wrapped class, rather than the base
    C++ _otio modules.
    """

    # where the python wrapped classes live
    SEARCH_MODULES = [
        otio.schema,
        otio.opentime,
        otio.core,
    ]

    # the C++ modules
    IGNORE_MODS = {
        otio._otio,
        otio._opentime
    }

    for mod in SEARCH_MODULES:
        result = _search_mod_recursively(cl, mod, IGNORE_MODS)
        if result is not None:
            return result

    return inspect.getmodule(cl).__name__


def _write_documentation(model):
    md_with_helpstrings = io.StringIO()
    md_only_fields = io.StringIO()

    md_with_helpstrings.write(DOCUMENT_HEADER)
    md_only_fields.write(FIELDS_ONLY_HEADER)

    modules = {}
    for cl in model:
        modobj = cl.__module__

        if modobj in ['opentimelineio._opentime', 'opentimelineio._otio']:
            modobj = _remap_to_python_modules(cl)

        modules.setdefault(modobj, []).append(cl)

    CURRENT_MODULE = None
    for module_list in sorted(modules):
        this_mod = ".".join(module_list.split('.')[:2])
        if this_mod != CURRENT_MODULE:
            CURRENT_MODULE = this_mod
            md_with_helpstrings.write(MODULE_HEADER.format(modname=this_mod))
            md_only_fields.write(MODULE_HEADER.format(modname=this_mod))

        # because these are classes, they need to sort on their stringified
        # names
        for cl in sorted(modules[module_list], key=lambda cl: str(cl)):
            modname = this_mod
            label = model[cl]["OTIO_SCHEMA"]

            if (cl.__doc__ is not None):
                docstring = inspect.getdoc(cl).split("\n")
                new_docstring = []
                for line in docstring:
                    line = textwrap.wrap(line, width=100,
                                         expand_tabs=False,
                                         replace_whitespace=False,
                                         drop_whitespace=False,
                                         break_long_words=False)
                    if (line == []):
                        line = [""]
                    for wrapped_line in line:
                        new_docstring.append(wrapped_line)
                new_docstring = "\n".join(new_docstring)
            else:
                new_docstring = cl.__doc__

            md_with_helpstrings.write(
                CLASS_HEADER_WITH_DOCS.format(
                    classname=label,
                    modpath=modname + "." + cl.__name__,
                    docstring=new_docstring
                )
            )
            md_only_fields.write(
                CLASS_HEADER_ONLY_FIELDS.format(
                    classname=label,
                )
            )

            for key, helpstr in sorted(model[cl].items()):
                if key in SKIP_KEYS:
                    continue
                md_with_helpstrings.write(
                    PROP_HEADER.format(propkey=key, prophelp=helpstr)
                )
                md_only_fields.write(
                    PROP_HEADER_NO_HELP.format(propkey=key)
                )

    return md_with_helpstrings.getvalue(), md_only_fields.getvalue()


def main():
    """  main entry point  """
    args = _parsed_args()
    with_docs, without_docs = generate_and_write_documentation()

    # print it out somewhere
    if args.dryrun:
        print(with_docs)
        return

    output = args.output
    if not output:
        output = tempfile.NamedTemporaryFile(
            'w',
            suffix="otio_serialized_schema.md",
            delete=False
        ).name

    with open(output, 'w') as fo:
        fo.write(with_docs)

    # write version without docstrings
    prefix, suffix = output.rsplit('.', 1)
    output_only_fields = prefix + "-only-fields." + suffix

    with open(output_only_fields, 'w') as fo:
        fo.write(without_docs)

    print(f"wrote documentation to {output} and {output_only_fields}")


def generate_and_write_documentation():
    model = _generate_model()
    return _write_documentation(model)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Print statistics about the otio file, including validation information."""

import argparse
import sys

import opentimelineio as otio


def _parsed_args():
    """ parse commandline arguments with argparse """

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        'filepath',
        type=str,
        nargs='+',
        help='files to operate on'
    )

    return parser.parse_args()


TESTS = []


def stat_check(name):
    def real_stat_check(fn):
        TESTS.append((name, fn))
        return fn
    return real_stat_check


@stat_check("parsed")
def _did_parse(input):
    return input and True or False


@stat_check("top level object")
def _top_level_object(input):
    return f"{input.schema_name()}.{input.schema_version()}"


@stat_check("number of tracks")
def _num_tracks(input):
    try:
        return len(input.tracks)
    except AttributeError:
        return 0


@stat_check("Tracks are the same length")
def _equal_length_tracks(tl):
    if not tl.tracks:
        return True
    for i, track in enumerate(tl.tracks):
        if track.duration() != tl.tracks[0].duration():
            raise RuntimeError(
                "track {} is not the same duration as the other tracks."
                " Track {} duration, vs: {}".format(
                    i,
                    track.duration(),
                    tl.tracks[0].duration()
                )
            )
    return True


@stat_check("deepest nesting")
def _deepest_nesting(input):
    def depth(parent):
        if not isinstance(parent, otio.core.Composition):
            return 1
        d = 0
        for child in parent:
            d = max(d, depth(child) + 1)
        return d
    if isinstance(input, otio.schema.Timeline):
        return depth(input.tracks) + 1
    else:
        return depth(input)


@stat_check("number of clips")
def _num_clips(input):
    return len(list(input.find_clips()))


@stat_check("total duration")
def _total_duration(input):
    try:
        return input.tracks.duration()
    except AttributeError:
        return "n/a"


@stat_check("total duration in timecode")
def _total_duration_timecode(input):
    try:
        d = input.tracks.duration()
        return otio.opentime.to_timecode(d, d.rate)
    except AttributeError:
        return "n/a"


@stat_check("top level rate")
def _top_level_rate(input):
    try:
        return input.tracks.duration().rate
    except AttributeError:
        return "n/a"


@stat_check("clips with cdl data")
def _clips_with_cdl_data(input):
    return len(list(c for c in input.find_clips() if 'cdl' in c.metadata))


@stat_check("Tracks with non standard types")
def _sequences_with_non_standard_types(input):
    return len(
        list(
            c
            for c in input.find_children(descended_from_type=otio.schema.Track)
            if c.kind not in (otio.schema.TrackKind.__dict__)
        )
    )


def _stat_otio(input_otio):
    for (test, testfunc) in TESTS:
        try:
            print(f"{test}: {testfunc(input_otio)}")
        except (otio.exceptions.OTIOError) as e:
            sys.stderr.write(
                "There was an OTIO Error: "
                " {}\n".format(e),
            )
            continue
        except (Exception) as e:
            sys.stderr.write(f"There was a system error: {e}\n")
            continue


def main():
    """  main entry point  """
    args = _parsed_args()

    for fp in args.filepath:
        try:
            parsed_otio = otio.adapters.read_from_file(fp)
        except (otio.exceptions.OTIOError) as e:
            sys.stderr.write(
                "The file did not successfully parse, with error:"
                " {}\n".format(e),
            )
            continue
        except (Exception) as e:
            sys.stderr.write(f"There was a system error: {e}\n")
            continue

        _stat_otio(parsed_otio)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Generates documentation of all the built in plugins for OpenTimelineIO"""

# @TODO: HookScripts and Hooks don't have formatters.  Because OTIO doesn't
#        ship with any, I've left places for them but haven't fleshed them out.

import argparse
import tempfile
import textwrap

import io

import opentimelineio as otio


# force using the same path separator regardless of what the OS says
# this ensures same behavior on windows and linux
PATH_SEP = "/"

ALL_PLUGINS_TEXT = """This documents all the plugins that otio could find."""

PUBLIC_ONLY_TEXT = """This documents all the plugins that ship with in the open
source OpenTimelineIO distribution.
"""

DOCUMENT_HEADER = """# Plugin Documentation

{plugin_header_text}

This document is automatically generated by running the
`autogen_plugin_documentation` command, or by running `make plugin-model`.  It
is part of the unit tests suite and should be updated whenever the schema
changes.  If it needs to be updated, run: `make doc-plugins-update` and this
file should be regenerated.

# Manifests

The manifests describe plugins that are visible to OpenTimelineIO.  The core
manifest is listed first, then any user-defined local plugins.

{manifests}

# Core Plugins

Manifest path: `{manifest_path}`

{manifest_contents}

{local_manifest_text}
"""

MANIFEST_CONTENT_TEMPLATE = """

## Adapter Plugins

Adapter plugins convert to and from OpenTimelineIO.

[Adapters documentation page for more information](./adapters).

[Tutorial on how to write an adapter](write-an-adapter).

{adapters}

## Media Linkers

Media Linkers run after the adapter has read in the file and convert the media
references into valid references where appropriate.

[Tutorial on how to write a Media Linker](write-a-media-linker).

{media_linkers}

## SchemaDefs

SchemaDef plugins define new external schema.

[Tutorial on how to write a schemadef](write-a-schemadef).

{schemadefs}

## HookScripts

HookScripts are extra plugins that run on _hooks_.

[Tutorial on how to write a hookscript](write-a-hookscript).

{hook_scripts}

## Hooks

Hooks are the points at which hookscripts will run.

{hooks}

"""

LOCAL_MANIFEST_TEMPLATE = """
# Local Manifests

Local manifests found:
{manifest_paths}

{local_manifest_body}
"""

PLUGIN_TEMPLATE = """
### {name}

```
{doc}
```

*source*: `{path}`

{other}

"""

ADAPTER_TEMPLATE = """
*Supported Features (with arguments)*:

{}

"""

SCHEMADEF_TEMPLATE = """
*Serializable Classes*:

{}
"""


def _parsed_args():
    """ parse commandline arguments with argparse """

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-d",
        "--dryrun",
        action="store_true",
        default=False,
        help="Dryrun mode - print out instead of perform actions"
    )
    group.add_argument(
        "-o",
        "--output",
        type=str,
        default=None,
        help="Update the baseline with the current version"
    )
    parser.add_argument(
        "-p",
        "--public-only",
        default=False,
        action="store_true",
        help=(
            "Only include plugins defined in the public open source manifests."
            "  Used by unit test."
        )
    )
    parser.add_argument(
        "-s",
        "--sanitized-paths",
        default=False,
        action="store_true",
        help="Sanitize paths to only show last three directories in a path."
    )

    return parser.parse_args()


def _format_plugin(plugin_map, extra_stuff, sanitized_paths):
    # XXX: always force unix path separator so that the output is consistent
    # between every platform.
    PATH_SEP = "/"

    path = plugin_map['path']

    # force using PATH_SEP in place of os.path.sep
    path = path.replace("\\", PATH_SEP)

    if sanitized_paths:
        path = PATH_SEP.join(path.split(PATH_SEP)[-3:])
    return PLUGIN_TEMPLATE.format(
        name=plugin_map['name'],
        doc=plugin_map['doc'],
        path=path,
        other=extra_stuff,
    )


def _format_doc(docstring, prefix):
    """Use textwrap to format a docstring for markdown."""

    initial_indent = prefix
    # subsequent_indent = " " * len(prefix)
    subsequent_indent = " " * 2

    block = docstring.split("\n")
    fmt_block = []
    for line in block:
        line = textwrap.fill(
            line,
            initial_indent=initial_indent,
            subsequent_indent=subsequent_indent,
            width=len(subsequent_indent) + 80,
        )
        initial_indent = subsequent_indent
        fmt_block.append(line)

    return "\n".join(fmt_block)


def _format_adapters(plugin_map):
    feature_lines = []

    for feature, feature_data in plugin_map['supported features'].items():
        doc = feature_data['doc']
        if doc:
            feature_lines.append(
                _format_doc(doc, f"- {feature}: \n```\n") + "\n```"
            )
        else:
            feature_lines.append(
                f"- {feature}:"
            )

        for arg in feature_data["args"]:
            feature_lines.append(f"  - {arg}")

    return ADAPTER_TEMPLATE.format("\n".join(feature_lines))


def _format_schemadefs(plugin_map):
    feature_lines = []

    for sd in plugin_map['SchemaDefs'].keys():
        doc = plugin_map['SchemaDefs'][sd]['doc']
        if doc:
            feature_lines.append(
                _format_doc(doc, f"- {sd}: \n```\n") + "\n```"
            )
        else:
            feature_lines.append(f"- {sd}:")

    return SCHEMADEF_TEMPLATE.format("\n".join(feature_lines))


_PLUGIN_FORMAT_MAP = {
    "adapters": _format_adapters,
    "schemadefs": _format_schemadefs,
}


def _manifest_formatted(
        plugin_info_map,
        manifest_paths=None,
        sanitized_paths=False
):
    display_map = {}

    for pt in otio.plugins.manifest.OTIO_PLUGIN_TYPES:
        pt_lines = []

        if pt == "hooks":
            # hooks get handled differently by plugin_info_map() so we will skip them
            display_map[pt] = ""
            continue

        sorted_plugins = [
            plugin_info_map[pt][name]
            for name in sorted(plugin_info_map[pt].keys())
        ]

        for plug in sorted_plugins:
            if "ERROR" in plug or not plug:
                continue

            # filter out plugins from other manifests
            if manifest_paths and plug['from manifest'] not in manifest_paths:
                continue

            plugin_stuff = ""
            try:
                plugin_stuff = _PLUGIN_FORMAT_MAP[pt](plug)
            except KeyError:
                pass

            plug_lines = _format_plugin(plug, plugin_stuff, sanitized_paths)

            pt_lines.append(plug_lines)

        display_map[pt] = "\n".join(str(line) for line in pt_lines)

    return MANIFEST_CONTENT_TEMPLATE.format(
        adapters=display_map['adapters'],
        media_linkers=display_map['media_linkers'],
        schemadefs=display_map['schemadefs'],
        hook_scripts=display_map['hook_scripts'],
        hooks=display_map['hooks'],
    )


def generate_and_write_documentation_plugins(
        public_only=False,
        sanitized_paths=False
):
    md_out = io.StringIO()

    plugin_info_map = otio.plugins.plugin_info_map()

    # start with the manifest list
    manifest_path_list = plugin_info_map['manifests'][:]

    if public_only:
        # keep only core manifests
        manifest_path_list = [
            p for p in manifest_path_list
            if p.replace("\\", PATH_SEP).split(PATH_SEP)[-3] == "opentimelineio"
        ]

    sanitized_path_list = manifest_path_list.copy()

    if sanitized_paths:
        # conform all paths to unix-style path separators and leave relative
        # paths (relative to root of OTIO directory)
        sanitized_path_list = [
            PATH_SEP.join(p.replace("\\", PATH_SEP).split(PATH_SEP)[-3:])
            for p in sanitized_path_list
        ]

    manifest_list = "\n".join(f"- `{mp}`" for mp in sanitized_path_list)

    core_manifest_path = manifest_path_list[0]
    core_manifest_path_sanitized = sanitized_path_list[0]
    core_manifest_text = _manifest_formatted(
        plugin_info_map,
        [core_manifest_path],
        sanitized_paths
    )

    local_manifest_text = ""
    if len(plugin_info_map) > 2 and not public_only:
        local_manifest_paths = manifest_path_list[2:]
        local_manifest_paths_sanitized = sanitized_path_list[2:]
        local_manifest_list = "\n".join(
            f"- `{mp}`" for mp in local_manifest_paths_sanitized
        )
        local_manifest_body = _manifest_formatted(
            plugin_info_map,
            local_manifest_paths,
            sanitized_paths
        )
        local_manifest_text = LOCAL_MANIFEST_TEMPLATE.format(
            manifest_paths=local_manifest_list,
            local_manifest_body=local_manifest_body,
        )

    header = PUBLIC_ONLY_TEXT if public_only else ALL_PLUGINS_TEXT

    md_out.write(
        DOCUMENT_HEADER.format(
            plugin_header_text=header,
            manifests=manifest_list,
            manifest_path=core_manifest_path_sanitized,
            manifest_contents=core_manifest_text,
            local_manifest_text=local_manifest_text,
        )
    )

    return md_out.getvalue()


def main():
    """  main entry point  """

    args = _parsed_args()

    docs = generate_and_write_documentation_plugins(
        args.public_only,
        args.sanitized_paths
    )

    # print it out somewhere
    if args.dryrun:
        print(docs)
        return

    output = args.output
    if not output:
        output = tempfile.NamedTemporaryFile(
            'w',
            suffix="otio_serialized_schema.md",
            delete=False
        ).name

    with open(output, 'w') as fo:
        fo.write(docs)

    print(f"wrote documentation to {output}.")


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

from .. core._core_utils import add_method
from .. import _otio


@add_method(_otio.Effect)
def __str__(self):
    return (
        "Effect("
        "{}, "
        "{}, "
        "{}, "
        "{}"
        ")".format(
            str(self.name),
            str(self.effect_name),
            str(self.metadata),
            str(self.enabled)
        )
    )


@add_method(_otio.Effect)
def __repr__(self):
    return (
        "otio.schema.Effect("
        "name={}, "
        "effect_name={}, "
        "metadata={}, "
        "enabled={}"
        ")".format(
            repr(self.name),
            repr(self.effect_name),
            repr(self.metadata),
            repr(self.enabled)
        )
    )



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Algorithms for OTIO objects."""

# flake8: noqa
from .track_algo import (
    track_trimmed_to_range,
    track_with_expanded_transitions
)

from .stack_algo import (
    flatten_stack,
    top_clip_at_time,
)

from .filter import (
    filtered_composition,
    filtered_with_sequence_context
)
from .timeline_algo import (
    timeline_trimmed_to_range
)
